from moai.data.iterator import Indexed
from moai.statistics import NoOp as NoStatistics

import moai.networks.lightning as minet

import torch
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["Presenter"]

class Presenter(minet.FeedForward):
    def __init__(self, 
        feedforward:        omegaconf.DictConfig=None,
        monads:             omegaconf.DictConfig=None,
        visualization:      omegaconf.DictConfig=None,
        data:               omegaconf.DictConfig=None,
        statistics:         omegaconf.DictConfig=None,
    ):        
        super(Presenter, self).__init__(
            feedforward=feedforward, data=data,
            monads=monads, visualization=visualization,            
        )
        self.statistics = hyu.instantiate(statistics) if statistics is not None else\
            NoStatistics()
        self.param = torch.nn.Linear(1, 1) # dummy layer
        self.global_test_step = 0

    def forward(self,
        x: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        return x

    def training_step(self, 
        batch:                  typing.Dict[str, torch.Tensor],
        batch_idx:              int,
        optimizer_idx:          int=None,
    ) -> typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]:
        preprocessed = self.preprocess(batch)
        prediction = self(preprocessed)
        postprocessed = self.postprocess(prediction)
        self.statistics(postprocessed)
        total_loss = torch.zeros(1, requires_grad=True)
        return { 'loss': total_loss, 'tensors': postprocessed }

    def training_step_end(self,
        train_outputs: typing.Dict[str, typing.Union[torch.Tensor, typing.Dict[str, torch.Tensor]]]
    ) -> None:
        if self.global_step and (self.global_step % self.visualization.interval == 0):
            self.visualization(train_outputs['tensors'])
        return train_outputs['loss']

    def validation_step(self,
        batch: typing.Dict[str, torch.Tensor],
        batch_nb: int
    ) -> dict:
        preprocessed = self.preprocess(batch)
        prediction = self(preprocessed)
        outputs = self.postprocess(prediction)
        self.statistics(outputs)
        return None

    def validation_epoch_end(self,
        outputs: typing.List[dict]
    ) -> dict:
        pass
    
    def test_step(self, 
        batch: typing.Dict[str, torch.Tensor],
        batch_nb: int
    ) -> dict:
        preprocessed = self.preprocess(batch)        
        prediction = self(preprocessed)
        outputs = self.postprocess(prediction)
        self.statistics(outputs)
        self.global_test_step += 1
        return torch.zeros(1), outputs

    def test_step_end(self,
        metrics_tensors: typing.Tuple[typing.Dict[str, torch.Tensor], typing.Dict[str, torch.Tensor]],        
    ) -> None:
        metrics, tensors = metrics_tensors
        if self.global_test_step and (self.global_test_step % self.exporter.interval == 0):
            self.exporter(tensors)
        if self.global_test_step and (self.global_test_step % self.visualization.interval == 0):
            self.visualization(tensors)
        return metrics

    def test_epoch_end(self, 
        outputs: typing.List[dict]
    ) -> dict:
        pass

    def train_dataloader(self) -> torch.utils.data.DataLoader:
        if hasattr(self.data.train.iterator, '_target_'):
            log.info(f"Instantiating ({self.data.train.iterator._target_.split('.')[-1]}) train set data iterator")
            train_iterator = hyu.instantiate(self.data.train.iterator)
        else:
            train_iterator = Indexed(
                self.data.train.iterator.datasets,
                self.data.train.iterator.augmentation if hasattr(self.data.train.iterator, 'augmentation') else None,
            )
        if not hasattr(self.data.train, 'loader'):
            log.error("Train data loader missing. Please add a data loader (i.e. \'- data/train/loader: torch\') entry in the configuration.")
        else:
            train_loader = hyu.instantiate(self.data.train.loader, train_iterator)
        return train_loader

    def val_dataloader(self) -> torch.utils.data.DataLoader:
        if hasattr(self.data.val.iterator, '_target_'):
            log.info(f"Instantiating ({self.data.val.iterator._target_.split('.')[-1]}) validation set data iterator")
            val_iterator = hyu.instantiate(self.data.val.iterator)
        else:
            val_iterator = Indexed(
                self.data.val.iterator.datasets,
                self.data.val.iterator.augmentation if hasattr(self.data.val.iterator, 'augmentation') else None,
            )
        if not hasattr(self.data.val, 'loader'):
            log.error("Validation data loader missing. Please add a data loader (i.e. \'- data/val/loader: torch\') entry in the configuration.")
        else:
            validation_loader = hyu.instantiate(self.data.val.loader, val_iterator)
        return validation_loader

    def test_dataloader(self) -> torch.utils.data.DataLoader:
        if hasattr(self.data.test.iterator, '_target_'):
            log.info(f"Instantiating ({self.data.test.iterator._target_.split('.')[-1]}) test set data iterator")
            test_iterator = hyu.instantiate(self.data.test.iterator)
        else:
            test_iterator = Indexed(
                self.data.test.iterator.datasets,
                self.data.test.iterator.augmentation if hasattr(self.data.test.iterator, 'augmentation') else None,
            )
        if not hasattr(self.data.test, 'loader'):
            log.error("Test data loader missing. Please add a data loader (i.e. \'- data/test/loader: torch\') entry in the configuration.")
        else:
            test_loader = hyu.instantiate(self.data.test.loader, test_iterator)
        return test_loader


from moai.utils.arguments import ensure_numeric_list

import random
import sys
import os
import hydra
import omegaconf.omegaconf
import hiplot as hip
import logging
import typing
import toolz
import yaml
import datetime
import parsedatetime as pdt
import glob
import pandas as pd
import tqdm
import functools

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

__DATE_TIME_FORMAT__ = "%Y-%m-%d-%H-%M-%S"

__DATE_TIME_CONVERTERS__ = {
    1: lambda d: datetime.datetime.combine(d, datetime.time()), # date
    2: lambda t: datetime.datetime.combine(datetime.date(), t), # time
    3: lambda dt: dt, # datetime
}

def parse_machine_date_time(string: str) -> datetime.datetime: 
    return datetime.datetime.strptime(string, __DATE_TIME_FORMAT__)

def parse_human_date_time(day: str, time: str) -> datetime.datetime:
    cal = pdt.Calendar()
    d = day if day else ''
    t = time if time else ''
    dt, result = cal.parseDT(f"{d} {t}".strip())
    if not result:
        log.error(f"Given day/time arguments ({day}/{time}) are not valid.")    
    return __DATE_TIME_CONVERTERS__[result](dt)

def get_conf(path: str) -> omegaconf.OmegaConf:
    if os.path.splitext(path)[1] != '.yaml':
        nominal_config_path = os.path.join(path, 'config_resolved.yaml')
        if os.path.exists(nominal_config_path):
            path = nominal_config_path
        elif os.path.basename(path) != '.hydra':
            path = os.path.join(path, '.hydra') #TODO: merge with overrides
            path = os.path.join(path, 'config.yaml')
    return path

def safe_get_dict(key: str, dictionary: dict):
    return toolz.get(key, dictionary, { })

def extract_losses(
    config:             omegaconf.DictConfig,
    path:               str,
): 
    result = { }
    supervision = toolz.get_in(["model", "supervision"], config)
    supervision = toolz.dissoc(supervision, "losses", "_target_")        
    for k, v in supervision.items():
        w = v['weight'] if 'weight' in v else 1.0
        w = sum(w) if isinstance(w, list) else w
        result[k] = w
    return result

def extract_metrics(
    moai_config:            omegaconf.DictConfig,
    path:                   str,
    options:                typing.Union[str, int],
):
    val_csv = glob.glob(os.path.join(path, "*_val.csv"))
    result = { }
    if val_csv:
        result = { }
        metrics = pd.read_csv(toolz.first(val_csv))
        metric_names = toolz.get_in(["model", "validation"], moai_config)
        metric_names = toolz.dissoc(metric_names, "metrics", "_target_", "indicators")
        if options.epoch_mode == 'best':
            best_metrics = { }
            for k, v in metric_names.items():
                name = v['out'] if 'out' in v else k
                name = os.path.commonprefix(name) if isinstance(name, list) else name
                best_metrics[name] = (metrics[name].max(), metrics['epoch'][metrics[name].idxmax()])
            min_epoch = int(min(map(lambda v: v[1], best_metrics.values())))        
            result['epoch'] = min_epoch
            for k, (v, e) in best_metrics.items():
                result[k] = v
                if e == min_epoch:
                    result['best'] = k
        else:
            epoch_value = [0, sys.maxsize] if options.epoch_mode == 'all' else ensure_numeric_list(options.epoch_value)
            min_epoch, max_epoch = tuple(epoch_value) if isinstance(epoch_value, omegaconf.ListConfig) and len(epoch_value) == 2 else (epoch_value[0], epoch_value[0])
            for i, e in enumerate(metrics['epoch']):
                e = int(e)
                if e >= min_epoch and e <= max_epoch:
                    result[i] = { }
                    result[i]['epoch'] = e
            for k, v in metric_names.items():
                name = v['out'] if 'out' in v else k
                name = os.path.commonprefix(name) if isinstance(name, list) else name
                for i, (e, v) in enumerate(zip(metrics['epoch'], metrics[name])):
                    e = int(e)
                    if e >= min_epoch and e <= max_epoch:
                        result[i][name] = v

    else:
        log.error(f"Could not find a validation metrics .csv file in {path}, skipping metrics for this experiment.")
    return result

def extract_optimization(
    moai_config:             omegaconf.DictConfig,
):
    result = { }
    optimizer = toolz.get_in(["model", "parameters", "optimization"], moai_config)
    learning_rates = { }
    lr = toolz.get("lr", optimizer, default=None)
    if lr:
        learning_rates[optimizer['_target_'].split('.')[-1]] = lr
    optimizers = toolz.get("optimizers", optimizer, default=None)
    if optimizers and isinstance(optimizers, dict):
        if len(optimizers) == 1 and "lr" in toolz.first(optimizers.values()):
            learning_rates[toolz.first(optimizers.keys())] = toolz.first(optimizers.values())["lr"]
        else:
            for optim, params in optimizers.items():
                if toolz.get("lr", params, default=None):
                    learning_rates[optim] = params["lr"]
    for k, v in learning_rates.items():
        result['optimizer'] = k
        result['lr'] = v
    return result

def extract_data(
    path: str, 
    plot_config: omegaconf.DictConfig
) -> typing.Mapping[str, typing.Union[str, float]]:
    data_frame = { }
    config_file = get_conf(path)
    with open(config_file, 'r') as f:
        moai_config = yaml.load(f, Loader=yaml.SafeLoader)
    if plot_config.losses:
        data_frame['losses'] = extract_losses(moai_config, path)
    if plot_config.metrics:
        data_frame['metrics'] = extract_metrics(moai_config, path, plot_config.metrics_options)
    if plot_config.optimizer: #TODO: add params & init func (i.e. pretrained)
        data_frame['optimization'] = extract_optimization(moai_config)
    if plot_config.monads:
        monads = toolz.get_in(["model", "monads"], moai_config)
        if plot_config.monads_options.key:
            data_frame['monads'] = { }
            for key in plot_config.monads_options.key:
                data_frame['monads'][key] = 'yes' if key in monads else 'no'
        if plot_config.monads_options.value:
            for k, v in plot_config.monads_options.value.items():
                data_frame['monads'][f"{k}_{v}"] = toolz.get_in([k, v], monads)
    #TODO: add model related data
    return data_frame

def post_process(
    data: typing.List[typing.Dict[str, typing.Dict[str, typing.Union[str, float, int]]]],
    config: omegaconf.DictConfig,
) -> typing.List[typing.Dict[str, typing.Union[str, float, int]]]:
    if config.losses:
        log.info("Adding losses ...")
        loss_keys = functools.reduce(lambda s, d: s | set(safe_get_dict('losses', d).keys()), data, set())
        losses_empty = dict(((k, 0.0) for k in loss_keys))
    if config.metrics:
        log.info("Adding metrics ...")
        metric_keys = functools.reduce(lambda s, d: s | set(safe_get_dict('metrics', d).keys()), data, set())
        metrics_empty = dict(((k, 0.0) for k in metric_keys))
    if config.monads:
        log.info("Adding monads ...")
        monads_keys = functools.reduce(lambda s, d: s | set(safe_get_dict('monads', d).keys()), data, set())
        monads_empty = dict(((k, 0.0) for k in monads_keys))
    consolidated_data = []
    for datum in data:
        res = { 'date': datum['date'] }
        if config.optimizer:
            res = toolz.merge(res, toolz.get('optimization', datum, { }))
        if config.losses:
            res = toolz.merge(res, losses_empty, toolz.get('losses', datum, { }))
        if config.monads:
            res = toolz.merge(res, monads_empty, toolz.get('monads', datum, { }))
        if config.metrics:
            if config.metrics_options.epoch_mode == 'best':
                res = toolz.merge(res, metrics_empty, toolz.get('metrics', datum, { }))
                consolidated_data.append(res)
            else:
                for row in datum['metrics'].values():
                    consolidated_data.append(toolz.merge(res, row))
    return consolidated_data

@hydra.main(config_name="tools/plot.yaml", config_path="conf")
def plot(cfg):
    from_date = parse_human_date_time(cfg.filter.date.start.day, cfg.filter.date.start.time)
    to_date = parse_human_date_time(cfg.filter.date.end.day, cfg.filter.date.end.time)
    if from_date > to_date:
        log.error("Invalid date ranges, start date ({from_date}) should be earlier than end date ({to_date}) date.")
        return
    log.info(f"Parsing {cfg.filter.name} experiments from {from_date} to {to_date}.")
    root = cfg.root
    date_paths = []
    root = root if os.path.isabs(root) else os.path.join(
        hydra.utils.get_original_cwd(), root
    )
    for date_folder in filter(
        lambda d: os.path.isdir(os.path.join(root, d)),
        os.listdir(root)
    ):
        for day_folder in filter(
            lambda d: os.path.isdir(os.path.join(root, date_folder, d)),
            os.listdir(os.path.join(root, date_folder))
        ):
            combined = f"{date_folder}-{day_folder}"[:19]
            date = parse_machine_date_time(combined)
            date_paths.append((combined, date, os.path.join(root, date_folder, day_folder)))
    paths = [(datetime, folder) for (datetime, date, folder) in date_paths if from_date <= date <= to_date]
    if not len(paths):
        log.error("Found no experiments")
        return
    log.info(f"Found {len(paths)} experiments.")
    data = []
    for path in tqdm.tqdm(paths, desc="Extracting data"):
        datum = extract_data(path[1], cfg)
        datum['date'] = path[0]
        data.append(datum)
    data = post_process(data, cfg)
    if data:
        hip.Experiment.from_iterable(data).to_html("plot.html")
        os.system("plot.html")
    else:
        log.error("No data collected.")

if __name__ == "__main__":
    # os.environ['HYDRA_FULL_ERROR'] = '1'
    sys.argv.append("hydra.run.dir=actions/${hydra.job.name}/${now:%Y-%m-%d}/${now:%H-%M-%S}")
    plot()



from moai.data.datasets.common import load_color_image
from moai.data.datasets.common.image2d import load_depth_image, load_mask_image
from moai.utils.arguments import ensure_string_list

import torch
import glob
import os
import typing
import logging
import toolz

__all__ = ["StructuredImages"]

log = logging.getLogger(__name__)

_LOADERS_ = {
    'color':        load_color_image,
    'image':        load_color_image,
    'depth':        load_depth_image,
    'range':        load_depth_image,
    'mask':         load_mask_image,
    'silhouette':   load_mask_image,
}

class StructuredImages(torch.utils.data.Dataset):
    def __init__(self,
        root:           str='',
        **kwargs:       typing.Mapping[str, typing.Mapping[str, typing.Any]],
    ):
        self.key_to_list = {}
        self.key_to_params = {}
        self.key_to_loader = {}
        for k, m in kwargs.items():
            glob_list = ensure_string_list(m['glob'])
            files = []
            for g in glob_list:
                files += glob.glob(os.path.join(root, g))
            self.key_to_list[k] = list(map(lambda f: os.path.join(root, f), files))
            self.key_to_params[k] = toolz.dissoc(m, 'type', 'glob')
            self.key_to_loader[k] = _LOADERS_[k] #m['type']

    def __len__(self) -> int:
        return len(next(iter(self.key_to_list.values())))

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        ret = { }
        for k, l in self.key_to_list.items():
            ret[k] = self.key_to_loader[k](l[index], **self.key_to_params[k])
        return ret


import moai.utils.engine as mieng

import torch
import omegaconf.omegaconf
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ['Optimizer']

class Optimizer(mieng.Single):
    def __init__(self,        
        parameters: typing.Iterator[torch.nn.Parameter],
        optimizers: omegaconf.DictConfig,
        **kwargs,
    ):
        super(Optimizer, self).__init__(
            items=optimizers, 
            name="optimizers",
            arguments=[parameters]
        )


import moai.nn.activation as mia
import moai.nn.convolution as mic
import moai.nn.residual as mires
import moai.nn.utils as miu
import moai.nn.sampling.spatial.downsample as mids
import moai.nn.sampling.spatial.upsample as mius

import torch
import omegaconf.omegaconf
import functools
import typing

#NOTE: implementation adapted from https://github.com/HRNet/HRNet-Facial-Landmark-Detection

__all__ = ["HighResolution"]

class HighResolution(torch.nn.Module):
    def __init__(self, 
        branches:              int,
        depth:                 int,
        start_features:        int,        
        fuse:                  omegaconf.DictConfig,
        residual:              omegaconf.DictConfig,
    ):
        super(HighResolution, self).__init__()
        self.num_branches = branches
        self.depth = depth
        self.branches = torch.nn.ModuleList([
            torch.nn.Sequential(*[
                mires.make_residual_block(
                    block_type=residual.type,
                    convolution_type=residual.convolution,
                    in_features=start_features * (2 ** b),
                    out_features=start_features * (2 ** b),
                    bottleneck_features=residual.bottleneck_features,
                    activation_type=residual.activation,
                    strided=False,
                    convolution_params={
                        'bias':         False,
                    }
                ) for d in range(self.depth)
            ]) for b in range(self.num_branches)
        ])
        self.fuse_layers = torch.nn.ModuleList([
            torch.nn.ModuleList([
                (torch.nn.Sequential(
                    *[
                        mic.make_conv_block(
                            block_type="conv2d",
                            convolution_type=fuse.convolution,
                            in_features=start_features * (2 ** j),
                            out_features=start_features * (2 ** i),
                            activation_type='none',
                            convolution_params={
                                'kernel_size':  1,
                                'stride':       1,
                                'padding':      0,
                                'bias':         False,
                            }),
                            mius.make_upsample(
                                upscale_type=fuse.upscale.type,
                                stride=2 ** (j-i),
                                features=start_features * (2 ** i),
                                scale=2 ** (j-i)
                            )
                    ][::(1 if fuse.upscale.conv_up else -1)]
                ) if j > i else ( # j < i -> j is lower res, we need to downscale
                    torch.nn.Sequential(*[
                        mic.make_conv_block(
                            block_type="conv2d",
                            convolution_type=fuse.convolution,
                            in_features=start_features * (2 ** j),
                            out_features=start_features * (2 ** i)\
                                if d == (i - j - 1) else start_features * (2 ** j),
                            activation_type=fuse.activation.prefusion\
                                if d == (i - j - 1) else fuse.activation.intermediate,
                            convolution_params={
                                'kernel_size':  3,
                                'stride':       2,
                                'padding':      1,
                                'bias':         False,
                            }) for d in range(i-j)
                        ]) if j < i else torch.nn.Identity()
                )) for j in range(self.num_branches) # each sequential operates on j scale
            ]) for i in range(self.num_branches) # each module has sequentials that result in i scale
        ])
        self.fuse_activ = mia.make_activation(
            activation_type=fuse.activation.final,
            features=sum(2 ** i for i in range(self.num_branches)),
            inplace=False,
        )

    def forward(self, 
        x: typing.Union[typing.List[torch.Tensor], typing.Tuple[torch.Tensor, ...]]
    ) -> typing.List[torch.Tensor]:
        for i, (b_i, x_i) in enumerate(zip(self.branches, x)):
            x[i] = b_i(x_i)
        for i, f in enumerate(self.fuse_layers):
            x[i] = self.fuse_activ(
                functools.reduce(
                    lambda s, y: s + y, 
                    (f_i(x_i) for f_i, x_i in zip(f, x))
                )
            )
        return x


from moai.data.datasets.common.image2d import (
    load_color_image,
    load_depth_image,
    load_mask_image,
    load_normal_image,
)
from moai.data.datasets.common.raw import load_npz_files

__all__ = [
    "load_color_image",
    "load_depth_image",
    "load_normal_image",
    "load_mask_image",
    "load_npz_files",
]


import moai.log.lightning as milog
import pytorch_lightning
import hydra.utils as hyu
import omegaconf.omegaconf
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["LightningPlayer"]

class LightningPlayer(pytorch_lightning.Trainer):
    def __init__(self,
        engine_callbacks:       typing.Sequence[pytorch_lightning.Callback]=[],
        logging:                omegaconf.DictConfig=None,
        num_nodes:              int=1,
        gpus:                   typing.Optional[typing.Union[typing.List[int], str, int]]=None,
        tpu_cores:              typing.Optional[int]=None,
        overfit_batches:        float=0.0,
        fast_dev_run:           bool=False,
        limit_train_batches:    float=1.0,
        limit_val_batches:      float=1.0,
        limit_test_batches:     float=1.0,
        val_check_interval:     float=1.0,
        #distributed_backend: typing.Optional[str]=None, NOTE: @PLT1.5
        weights_summary:        typing.Optional[str]='full',
        amp_level:              str='O2',
        **kwargs
    ):
        logger = hyu.instantiate(logging) if logging is not None else milog.NoOp()
        super(LightningPlayer, self).__init__(
            callbacks=engine_callbacks,
            logger=logger,      
            num_nodes=num_nodes,
            gpus=gpus,
            tpu_cores=tpu_cores,
            overfit_batches=overfit_batches,
            fast_dev_run=fast_dev_run,
            max_epochs=1,
            min_epochs=1,
            limit_train_batches=limit_train_batches,
            limit_val_batches=limit_val_batches,
            limit_test_batches=limit_test_batches,
            val_check_interval=val_check_interval,
            #distributed_backend=distributed_backend,
            weights_summary=weights_summary,
            amp_level=amp_level,
            **kwargs
        )

    def play(self, model):
        self.fit(model)
        log.info("Train & validation data completed, continuing with test data.")
        self.test(model)


from moai.monads.utils.common import (
    expand_dims,
    dim_list,
)
from moai.monads.utils.spatial import (
    expand_spatial_dims,
    spatial_dim_list,
    flatten_spatial_dims,
    spatial_dims
)

__all__ = [
    "expand_dims",
    "dim_list",
    "spatial_dim_list",
    "expand_spatial_dims",
    "flatten_spatial_dims",
    "spatial_dims"
]


import moai.nn.deconvolution as mid
import moai.nn.activation as mia

import torch

__all__ = [
    "Deconv2dBlock",
]

class Deconv2dBlock(torch.nn.Module):
    def __init__(self,
        deconvolution_type:str,
        activation_type: str,        
        in_features: int,
        out_features: int,
        deconvolution_params: dict,
        activation_params: dict,        
        **kwargs: dict 
    ):
        super(Deconv2dBlock, self).__init__()
        self.deconv = mid.make_deconv_op(
            deconvolution_type=deconvolution_type,
            in_channels=in_features,
            out_channels=out_features,
            **deconvolution_params
        )
        self.activation = mia.make_activation(
            features=out_features,
            activation_type=activation_type,
            **activation_params
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.activation(self.deconv(x))


from collections.abc import Callable

import torch
import functools
import logging

log = logging.getLogger(__name__)

try:
    from mapped_convolution.nn import MappedConvolution, MappedTransposedConvolution
except ImportError:
    __SPHERICAL_MAPPED_INCLUDED__ = False
else:
    __SPHERICAL_MAPPED_INCLUDED__ = True

__all__ = ["Kaiming"]

class Kaiming(Callable):
    def __init__(self, 
        zero_bias:      bool=False,
        gain:           float=0.0,
        mode:           str='fan_out', # one of ['fan_in', 'fan_out']
        nonlinearity:   str='relu', # one of ['relu', 'leaky_relu']
    ):
        self.conv_w_init = functools.partial(torch.nn.init.kaiming_normal_,
            a=gain, 
            mode=mode,
            nonlinearity=nonlinearity
        )
        self.conv_b_init = torch.nn.init.zeros_ if zero_bias else torch.nn.init.normal_
        self.w_init = torch.nn.init.normal_

    def __call__(self,
        module: torch.nn.Module
    ) -> None:        
        if isinstance(module, torch.nn.Conv1d):
            self.w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.Conv2d):
            self.conv_w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.Conv3d):
            self.conv_w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.ConvTranspose1d):
            self.w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.ConvTranspose2d):
            self.w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.ConvTranspose3d):
            self.w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.BatchNorm1d):
            torch.nn.init.normal_(module.weight, mean=1, std=0.02)
            torch.nn.init.constant_(module.bias, 0)
        elif isinstance(module, torch.nn.BatchNorm2d):
            torch.nn.init.normal_(module.weight, mean=1, std=0.02)
            torch.nn.init.constant_(module.bias, 0)
        elif isinstance(module, torch.nn.BatchNorm3d):
            torch.nn.init.normal_(module.weight, mean=1, std=0.02)
            torch.nn.init.constant_(module.bias, 0)
        elif isinstance(module, torch.nn.Linear):
            self.conv_w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif __SPHERICAL_MAPPED_INCLUDED__ and \
            (isinstance(module, MappedConvolution) or isinstance(module, MappedTransposedConvolution)):
                self.conv_w_init(module.weight)
                if module.bias is not None:
                    self.conv_b_init(module.bias)



import torch
import typing
import toolz
import numpy as np


def load_npz_files(
    filename:   str,
) -> typing.Dict[str, torch.Tensor]:
    data = np.load(filename)
    return toolz.valmap(lambda a: torch.from_numpy(a), data)



import moai.utils.engine as mieng

import omegaconf.omegaconf
import logging

log = logging.getLogger(__name__)

__all__ = ["Visualizer"]

class Visualizer(mieng.Interval, mieng.Single):
    def __init__(self,
        batch_interval:int,
        visualizers: omegaconf.DictConfig,        
    ):
        mieng.Interval.__init__(self, batch_interval)
        mieng.Single.__init__(self, items=visualizers)


import moai.nn.utils as miu

import torch
import functools
import logging
import typing

log = logging.getLogger(__name__)

__all__ = [
    "make_upsample",
]

__UPSAMPLE_FACTORY__ = {
    "none":                 torch.nn.Identity,
}

#NOTE: need to implement:
#  1) upscale->project,
#  2) project->upscale,
#  3) transpose conv with spatial and feature size reduction

def _update_upsample_op(name: str, type: typing.Type):
    if name not in __UPSAMPLE_FACTORY__.keys():
        __UPSAMPLE_FACTORY__.update({name: type})
    else:
        log.error(f"Trying to add an already existing key {name} in the convolution operation factory.")
 
def make_upsample(
    upscale_type: str,
    features: int,
    kernel_size: int=4,
    stride: int=2,
    **kwargs
) -> torch.nn.Module:
    if upscale_type in __UPSAMPLE_FACTORY__.keys():
        return miu.instantiate(__UPSAMPLE_FACTORY__[upscale_type], 
        **{
            **locals(),
            **kwargs 
        })
    else:
        log.error(f"Upscale type ({upscale_type}) not found.")

import moai.nn.sampling.spatial.upsample.deconv as mids

if "deconv2d" not in __UPSAMPLE_FACTORY__.keys():
    _update_upsample_op("deconv2d", mids.StridedDeconv2d)

del mids

import moai.nn.sampling.spatial.upsample.interpolate as miup

if "upsample2d" not in __UPSAMPLE_FACTORY__.keys():
    _update_upsample_op("upsample2d", miup.Upsample2d)

del miup


import torch
import functools
import typing

def expand_dims(
    src:            torch.Tensor,
    dst:            torch.Tensor,
    start_index:    int=1,
) -> torch.Tensor:
    r"""
        Expands the source tensor to match the spatial dimensions of the destination tensor.
        
        Arguments:
            src (torch.Tensor): A tensor of [B, K, X(Y)(Z)] dimensions
            dst (torch.Tensor): A tensor of [B, X(Y)(Z), (D), (H), W] dimensions
            start_index (int, optional): An optional start index denoting the start of the spatial dimensions
        
        Returns:
            A torch.Tensor of [B, K, X(Y)(Z), (1), (1), 1] dimensions. 
    """
    return functools.reduce(
        lambda s, _: s.unsqueeze(-1), 
        [*dst.shape[start_index:]],
        src
    )

def dim_list(
    tensor:         torch.Tensor,
    start_index:    int=1,
) -> typing.List[int]:
    return list(range(start_index, len(tensor.shape)))

def dims(
    tensor:         torch.Tensor,
    start_index:    int=1
) -> torch.Tensor:
    return torch.Tensor([tensor.size()[start_index:]]).squeeze().to(tensor.device)


import torch
import typing
import omegaconf.omegaconf
import functools

__all__ = [
    "Scalar",
    "Random",
    "Ones",
    "Zeros",
    "Clone",
    "Parameter",
    "Parameters",
]

class Scalar(torch.nn.Module):
    def __init__(self, 
        value: float
    ):
        super(Scalar, self).__init__()
        self.value = value
    
    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return torch.scalar_tensor(self.value,
            dtype=tensor.dtype, device=tensor.device
        )
torch.distributions.Normal
class Random(torch.nn.Module):
    __RANDOMS__ = {
        'unit': torch.rand,
        'normal': torch.randn,
    }

    def __init__(self,
        shape:      typing.Sequence[int],
        scale:      float=1.0,
        mode:       str='unit', # one of [unit, normal]
    ):
        super(Random, self).__init__()
        self.generate = Random.__RANDOMS__[mode]
        self.shape = shape
        self.scale = scale

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        b = tensor.shape[0]
        generated = self.generate([b, *self.shape], device=tensor.device)
        return generated * self.scale if self.scale != 1.0 else generated

class Ones(torch.nn.Module):
    def __init__(self):
        super(Ones, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        # return torch.ones_like(tensor)
        return torch.ones(1, *tensor.shape[1:], dtype=tensor.dtype,
                device=tensor.device).expand_as(tensor) if tensor.shape\
            else torch.scalar_tensor(1, dtype=tensor.dtype, device=tensor.device)

class Zeros(torch.nn.Module):
    def __init__(self):
        super(Zeros, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        # return torch.zeros_like(tensor)
        return torch.zeros(1, *tensor.shape[1:], 
            dtype=tensor.dtype, device=tensor.device).expand_as(tensor)

class Clone(torch.nn.Module):
    def __init__(self):
        super(Clone, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return tensor.clone()

class Parameter(torch.nn.Module):
    def __init__(self,
        shape:          typing.Union[int, typing.Sequence[int]],
        init:           str='zeros', # one of [zeros, ones, rand, randn],
    ):
        super(Parameter, self).__init__()
        self.register_parameter('value', torch.nn.Parameter(
            getattr(torch, init)(tuple(shape))) #TODO: check omegaconf's convert type annotation
        )

    def forward(self, void: torch.Tensor) -> torch.nn.parameter.Parameter:
        return self.value

class Parameters(torch.nn.Module):
    def __init__(self,
        parameters:         omegaconf.DictConfig,
    ):
        super(Parameters, self).__init__()
        for name, param in parameters.items():
            self.register_parameter(str(name), torch.nn.Parameter(
                getattr(torch, param.init or 'zeros')(tuple(param.shape))) #TODO: check omegaconf's convert type annotation
            )

    def forward(self, void: torch.Tensor) -> torch.nn.parameter.Parameter:
        return dict(self.named_parameters())



import moai.utils.engine as mieng

import torch.optim
import omegaconf.omegaconf
import typing
import logging

log = logging.getLogger(__name__)

__all__ = ["Scheduler"]

class Scheduler(mieng.Single):
    def __init__(self,
        optimizers: typing.Sequence[torch.optim.Optimizer],
        schedulers: omegaconf.DictConfig,        
    ):
        super(Scheduler, self).__init__(items=schedulers, 
            arguments=optimizers, name="schedulers"
        )


from moai.visualization.visdom.base import Base
from moai.utils.arguments import assert_numeric
from moai.monads.execution.cascade import _create_accessor

from matplotlib import pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure

import torch
import seaborn
import typing
import logging
import numpy as np
import pytorch_lightning

log = logging.getLogger(__name__)

__all__ = ["Density2d"]

class Density2d(Base, pytorch_lightning.Callback):
    def __init__(self,
        key:                typing.Union[str, typing.Sequence[str]],
        fill:               bool=True,
        palette:            str='coral',
        levels:             int=10,
        width:              int=720,
        height:             int=480,
        name:               str="default",
        ip:                 str="http://localhost",
        port:               int=8097,
    ):
        super().__init__(name, ip, port)
        self.levels = levels
        self.fill = fill
        self.palette = palette
        self.width = width / 300.0
        self.height = height / 300.0
        self.names = [key] if isinstance(key, str) else list(key)
        self.keys = [_create_accessor(k) for k in self.names]
        self.cache = {}

    @property
    def name(self) -> str:
        return self.env_name
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for n, k in zip(self.names, self.keys):
            self.cache[n] = k(tensors)
            
    def on_train_epoch_end(self, 
        trainer: pytorch_lightning.Trainer,
        pl_module: pytorch_lightning.LightningModule,
    ) -> None:
        for n in self.names:
            fig = Figure(figsize=(self.width, self.height), dpi=300.0)#self.dpi)
            canvas = FigureCanvasAgg(fig)
            ax = fig.add_subplot(111)
            seaborn.kdeplot(
                x=self.cache[n][:, 0], 
                y=self.cache[n][:, 1],
                fill=self.fill, palette=self.palette,
                ax=ax, levels=self.levels,
            )
            canvas.draw()
            img = np.asarray(canvas.buffer_rgba())
            self.visualizer.images(
                img[np.newaxis, ..., :3].transpose(0, 3, 1, 2),
                win=n, env=self.name
            )
            


from collections.abc import Callable
from moai.utils.arguments import (
    ensure_path,
    ensure_choices,
)

import moai.utils.color.colorize as mic

import os
import functools
import numpy as np
import torch
import torchvision
import typing
import logging
import cv2

__all__ = ["Image2d"]

log = logging.getLogger(__name__)

class Image2d(Callable):

    __MODES__ = ['overwrite', 'all']
    __FORMATS__ = ['jpg', 'png', 'exr']

    def __init__(self,
        path:       str,
        image:      typing.Union[str, typing.Sequence[str]],
        type:       typing.Union[str, typing.Sequence[str]],
        colormap:   typing.Union[str, typing.Sequence[str]],
        transform:  typing.Union[str, typing.Sequence[str]],
        mode:       str="overwrite", # all"
        extension:  typing.Union[str, typing.Sequence[str]]=["png"], # jpg or png or exr
    ):
        self.mode = ensure_choices(log, "saving mode", mode, Image2d.__MODES__)
        folder = ensure_path(log, "output folder", path)
        self.formats = [ensure_choices(log, "output format", ext, Image2d.__FORMATS__) for ext in extension]
        self.mode = mode
        self.index = 0
        self.keys = [image] if isinstance(image, str) else list(image)
        self.types = [type] if isinstance(type, str) else list(type)
        self.transforms = [transform] if isinstance(transform, str) else list(transform)
        self.colormaps = [colormap] if isinstance(colormap, str) else list(colormap)
        self.save_map = {
            'color': functools.partial(self._save_color, os.path.join(os.getcwd(), folder)),
            'depth': functools.partial(self._save_depth, os.path.join(os.getcwd(), folder)),
        }
        self.transform_map = {
            'none': functools.partial(self._no_transform),
            'minmax': functools.partial(self._minmax_normalization),
            "positive_minmax": functools.partial(self._minmax_normalization, positive_only=True),
            "log_minmax": functools.partial(self._log_minmax_normalization),
        }
        self.colorize_map = { "none": lambda x: x.cpu().numpy() }
        self.colorize_map.update(mic.COLORMAPS)
        
    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:
        for k, t, tf, c, f in zip(
            self.keys,
            self.types,
            self.transforms,
            self.colormaps,
            self.formats,
        ):
            b = self.save_map[t](
                self.colorize_map[c](
                    self.transform_map[tf](tensors[k])
                ), k, self.index, f
            )
        self.index = 0 if self.mode == "overwrite" else self.index + b

    @staticmethod
    def _save_color(
        path:               str,
        array:              np.array,
        key:                str,
        index:              int,
        ext:                str,
    ) -> int:
        b, _, __, ___ = array.shape
        for i in range(b):
        #    torchvision.utils.save_image(
        #        torch.from_numpy(array)[i, :, :, :], 
        #        f'{key}_{index + i}.{ext}'
        #     )
            cv2.imwrite(
                f'{key}_{index + i}.{ext}',
                array[i, :, :, :].transpose(1, 2, 0),
            )
        return b

    @staticmethod
    def _save_depth(
        path:               str,
        array:              np.array,
        key:                str,
        index:              int,
        ext:                str,
    ) -> int:
        b, _, __, ___ = array.shape
        for i in range(b): #NOTE: change to cv2 to support .exr
            cv2.imwrite(
                f'{key}_{index + i}.{ext}',
                array[i, :, :, :].transpose(1, 2, 0),
            )
        #    torchvision.utils.save_image(
        #        torch.from_numpy(array)[i, :, :, :], 
        #        f'{key}_{index + i}.{ext}'
        #     )
        return b

    @staticmethod #TODO: refactor these into a common module
    def _no_transform(tensor: torch.Tensor) -> torch.Tensor:
        return tensor

    @staticmethod #TODO: refactor these into a common module
    def _minmax_normalization(tensor: torch.Tensor, positive_only=False) -> torch.Tensor:
        b, _, __, ___ = tensor.size()
        t = tensor
        if positive_only:
            t[t < 0.0] = 0.0
        min_v = torch.min(t.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        max_v = torch.max(t.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        return (t - min_v) / (max_v - min_v)        
    
    @staticmethod #TODO: refactor these into a common module
    def _log_minmax_normalization(tensor: torch.Tensor) -> torch.Tensor:
        b, _, __, ___ = tensor.size()
        log_tensor = torch.log(tensor + 1e-8)
        min_v = torch.min(log_tensor.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        max_v = torch.max(log_tensor.view(b, -1), dim=1, keepdim=True)[0].unsqueeze(2).unsqueeze(3)
        return (log_tensor - min_v) / (max_v - min_v)


import torch

class Projection(torch.nn.Module):

    __XYZ_AT__ = ['channel', 'row']
    __UV_AT__ = ['channel', 'row']
    __UV_TYPE__ = ['ndc', 'coord', 'norm']

    def __init__(self,
        xyz_at:         str='channel',  # ['channel', 'row]
        uv_at:          str='channel',  # ['channel', 'row]
        uv_type:        str='ndc',      # ['ndc', 'coord', 'norm']
    ):
        super(Projection, self).__init__()
        self.homogemize = self._homogemize_channel if xyz_at == 'channel'\
            else self._homogenize_row
        self.format_uv = self._format_uv_channel if uv_at == 'channel'\
            else self._format_uv_row
        self.convert_uv = self._convert_ndc if uv_type == 'ndc'\
            else (self._convert_norm if uv_type == 'norm'\
                else lambda x , h , w: x[:,:2,...]
            )

    def _homogemize_channel(self,
        points:             torch.Tensor,
        epsilon:            float=1e-8,
    ) -> torch.Tensor:
        return points / (points[:, -1:, ...] + epsilon)

    def _homogenize_row(self, 
        points:             torch.Tensor,
        epsilon:            float=1e-8,
    ) -> torch.Tensor:
        x = points[..., 0]
        y = points[..., 1]
        z = points[..., 2] + epsilon
        x_h = x / z
        y_h = y / z
        ones = torch.ones_like(z)
        return torch.stack([x_h, y_h, ones], dim=1)

    def _convert_norm(self, 
        uvs:                torch.Tensor, 
        height:             int,
        width:              int,
    ) -> torch.Tensor:
        dims = torch.stack([
            torch.Tensor([[width]]), torch.Tensor([[height]])
        ], dim=1).to(uvs)
        return uvs[:, :2, ...] / dims

    def _convert_ndc(self, 
        uvs:                torch.Tensor, 
        height:             int,
        width:              int,
    ) -> torch.Tensor:
        one = self._convert_one(uvs, height, width)
        return one * 2.0 - 1.0

    def _format_uv_channel(self, 
        uvs:                torch.Tensor, 
        height:             int,
        width:              int,
    ) -> torch.Tensor:
        b = uvs.shape[0]
        dims = [b, 2, height, width] if height > 1 else [b, 2, width]
        return uvs.reshape(*dims)

    def _format_uv_row(self, 
        uvs:                torch.Tensor,
        height:             int,
        width:              int,
    ) -> torch.Tensor:
        uv = uvs.transpose(-1, -2)
        b = uv.shape[0]
        dims = [b, height, width, 2] if height > 1 else [b, width, 2]
        return uv.reshape(*dims)

    def forward(self,
        intrinsics:         torch.Tensor,
        points:             torch.Tensor,
        grid:               torch.Tensor=None,
    ) -> torch.Tensor:
        b = points.shape[0]        
        homogeneous = self.homogemize(points)
        uvs = intrinsics @ homogeneous.view(b, 3, -1)
        grid = points if grid is None else grid
        w = grid.shape[-1]
        h = grid.shape[-2] 
        uvs = self.convert_uv(uvs, h, w)
        w_p = uvs.shape[-1]
        h_p = uvs.shape[-2] if len(uvs.shape) > 3 else 0
        return self.format_uv(uvs, h_p, w_p)




from collections.abc import Callable

import torch
import logging

log = logging.getLogger(__name__)

__all__ = ["Xavier"]

class Xavier(Callable):
    def __init__(self,
        zero_bias: bool,
    ):        
        self.conv_w_init = torch.nn.init.xavier_normal_
        self.w_init = torch.nn.init.normal_
        self.conv_b_init = torch.nn.init.zeros_ if zero_bias else torch.nn.init.normal_

    def __call__(self, 
        module: torch.nn.Module
    ) -> None:
        if isinstance(module, torch.nn.Conv1d):
            self.w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.Conv2d):
            self.conv_w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.Conv3d):
            self.conv_w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.ConvTranspose1d):
            self.w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.ConvTranspose2d):
            self.conv_w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.ConvTranspose3d):
            self.conv_w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)
        elif isinstance(module, torch.nn.BatchNorm1d):
            self.w_init(module.weight, mean=1, std=0.02)
            torch.nn.init.constant_(module.bias, 0)
        elif isinstance(module, torch.nn.BatchNorm2d):
            self.w_init(module.weight, mean=1, std=0.02)
            torch.nn.init.constant_(module.bias, 0)
        elif isinstance(module, torch.nn.BatchNorm3d):
            self.w_init(module.weight, mean=1, std=0.02)
            torch.nn.init.constant_(module.bias, 0)
        elif isinstance(module, torch.nn.Linear):
            self.conv_w_init(module.weight)
            if module.bias is not None:
                self.conv_b_init(module.bias)



import torch
import cv2
import numpy as np
import logging
import typing

# NOTE: https://github.com/BoChenYS/BPnP
# NOTE: https://arxiv.org/pdf/1909.06043.pdf 


log = logging.getLogger(__name__)

try:
    import kornia as kn
except ImportError:
    log.error("Kornia is required for BPnP, please install it before proceeding.")

__all__ = ["BPnP"]

def _batch_project(P, pts3d, K, angle_axis=True):
    n = pts3d.size(0)
    bs = P.size(0)
    device = P.device
    pts3d_h = torch.cat((pts3d, torch.ones(n, 1, device=device)), dim=-1)
    if angle_axis:
        R_out = kn.angle_axis_to_rotation_matrix(P[:, 0:3].reshape(bs, 3))
        PM = torch.cat((R_out[:, 0:3, 0:3], P[:, 3:6].reshape(bs, 3, 1)), dim=-1)
    else:
        PM = P
    pts3d_cam = pts3d_h.matmul(PM.transpose(-2, -1))
    pts2d_proj = pts3d_cam.matmul(K.t())
    S = pts2d_proj[:,:, 2].reshape(bs, n, 1)
    pts2d_pro = pts2d_proj[:, :, 0:2] / (S + 1e-12)
    return pts2d_pro

def _get_coefs(P_6d, pts3d, K):
    device = P_6d.device
    n = pts3d.size(0)
    m = P_6d.size(-1)
    coefs = torch.zeros(n, 2, m, device=device)
    torch.set_grad_enabled(True)
    y = P_6d.repeat(n, 1)
    proj = _batch_project(y, pts3d, K).squeeze()
    vec = torch.diag(torch.ones(n,device=device).float())
    for k in range(2):
        torch.set_grad_enabled(True)
        y_grad = torch.autograd.grad(proj[:, :, k], y, vec, retain_graph=True, create_graph=True)
        coefs[:, k, :] = -2.0 * y_grad[0].clone()
    return coefs

class BPnPFunction_fast(torch.autograd.Function):
    """
    BPnP_fast is the efficient version of the BPnP class which ignores the higher order dirivatives through the coefs' graph. This sacrifices
    negligible gradient accuracy yet saves significant runtime. 
    INPUTS:
    pts2d - the 2D keypoints coordinates of size [batch_size, num_keypoints, 2]
    pts3d - the 3D keypoints coordinates of size [num_keypoints, 3]
    K     - the camera intrinsic matrix of size [3, 3]
    OUTPUT:
    P_6d  - the 6 DOF poses of size [batch_size, 6], where the first 3 elements of each row are the angle-axis rotation 
    vector (Euler vector) and the last 3 elements are the translation vector. 
    NOTE:
    This BPnP function assumes that all sets of 2D points in the mini-batch correspond to one common set of 3D points. 
    For situations where pts3d is also a mini-batch, use the BPnP_m3d class.
    """
    @staticmethod
    def forward(ctx, pts2d, pts3d, K, ini_pose=None):
        bs = pts2d.size(0)
        n = pts2d.size(1)
        device = pts2d.device
        K_np = np.array(K.detach().cpu())
        P_6d = torch.zeros(bs,6,device=device)

        for i in range(bs):
            pts2d_i_np = np.ascontiguousarray(pts2d[i].detach().cpu()).reshape((n,1,2))
            pts3d_i_np = np.ascontiguousarray(pts3d[i].detach().cpu()).reshape((n,3))
            if ini_pose is None:
                _, rvec0, T0 = cv2.solvePnP(
                    objectPoints=pts3d_i_np, 
                    imagePoints=pts2d_i_np, 
                    cameraMatrix=K_np, 
                    distCoeffs=None,
                    flags=cv2.SOLVEPNP_ITERATIVE, 
                    useExtrinsicGuess=False)
            else:
                rvec0 = np.array(ini_pose[i, 0:3].cpu().reshape(3, 1))
                T0 = np.array(ini_pose[i, 3:6].cpu().reshape(3, 1))
            _, rvec, T = cv2.solvePnP(objectPoints=pts3d_i_np, imagePoints=pts2d_i_np, cameraMatrix=K_np, distCoeffs=None, flags=cv2.SOLVEPNP_ITERATIVE, useExtrinsicGuess=True, rvec=rvec0, tvec=T0)
            _, rvec, T = cv2.solvePnP(objectPoints=pts3d_i_np, imagePoints=pts2d_i_np, cameraMatrix=K_np, distCoeffs=None, flags=cv2.SOLVEPNP_ITERATIVE, useExtrinsicGuess=True, rvec=rvec0, tvec=T0)
            angle_axis = torch.tensor(rvec,device=device,dtype=torch.float).reshape(1, 3)
            T = torch.tensor(T,device=device,dtype=torch.float).reshape(1, 3)
            P_6d[i,:] = torch.cat((angle_axis,T),dim=-1)

        ctx.save_for_backward(pts2d,P_6d,pts3d,K)
        return P_6d

    @staticmethod
    def backward(ctx, grad_output):
        pts2d, P_6d, pts3d, K = ctx.saved_tensors
        device = pts2d.device
        bs = pts2d.size(0)
        n = pts2d.size(1)
        m = 6

        grad_x = torch.zeros_like(pts2d)
        grad_z = torch.zeros_like(pts3d)
        grad_K = torch.zeros_like(K)

        for i in range(bs):
            J_fy = torch.zeros(m,m, device=device)
            J_fx = torch.zeros(m,2*n, device=device)
            J_fz = torch.zeros(m,3*n, device=device)
            J_fK = torch.zeros(m, 9, device=device)

            pts2d_flat = pts2d[i].clone().reshape(-1).detach().requires_grad_()
            P_6d_flat = P_6d[i].clone().reshape(-1).detach().requires_grad_()
            pts3d_flat = pts3d[i].clone().reshape(-1).detach().requires_grad_()
            coefs = _get_coefs(P_6d[i].reshape(1,6), pts3d_flat.reshape(n,3), K).detach()
            K_flat = K.clone().reshape(-1).detach().requires_grad_()

            for j in range(m):
                torch.set_grad_enabled(True)
                if j > 0:
                    pts2d_flat.grad.zero_()
                    P_6d_flat.grad.zero_()
                    pts3d_flat.grad.zero_()
                    K_flat.grad.zero_()

                R = kn.angle_axis_to_rotation_matrix(P_6d_flat[0:m-3].reshape(1,3))

                P = torch.cat((R[0,0:3,0:3].reshape(3,3), P_6d_flat[m-3:m].reshape(3,1)),dim=-1)
                KP = torch.mm(K_flat.reshape(3,3), P)
                pts2d_i = pts2d_flat.reshape(n,2).transpose(0,1)
                pts3d_i = torch.cat((pts3d_flat.reshape(n,3),torch.ones(n,1,device=device)),dim=-1).t()
                proj_i = KP.mm(pts3d_i)
                Si = proj_i[2,:].reshape(1,n)

                r = pts2d_i*Si-proj_i[0:2,:]
                coef = coefs[:,:,j].transpose(0,1) # size: [2,n]
                fj = (coef*r).sum()
                fj.backward()
                J_fy[j,:] = P_6d_flat.grad.clone()
                J_fx[j,:] = pts2d_flat.grad.clone()
                J_fz[j,:] = pts3d_flat.grad.clone()
                J_fK[j,:] = K_flat.grad.clone()

            inv_J_fy = torch.inverse(J_fy)

            J_yx = (-1) * torch.mm(inv_J_fy, J_fx)
            J_yz = (-1) * torch.mm(inv_J_fy, J_fz)
            J_yK = (-1) * torch.mm(inv_J_fy, J_fK)

            grad_x[i] = grad_output[i].reshape(1,m).mm(J_yx).reshape(n,2)
            grad_z += grad_output[i].reshape(1,m).mm(J_yz).reshape(n,3)
            grad_K += grad_output[i].reshape(1,m).mm(J_yK).reshape(3,3)

        return grad_x, grad_z, grad_K, None

class BPnPFunction_m3d(torch.autograd.Function):
    """
    BPnP_m3d supports mini-batch inputs of 3D keypoints, where the i-th set of 2D keypoints correspond to the i-th set of 3D keypoints.
    INPUTS:
    pts2d - the 2D keypoints coordinates of size [batch_size, num_keypoints, 2]
    pts3d - the 3D keypoints coordinates of size [batch_size, num_keypoints, 3]
    K     - the camera intrinsic matrix of size [3, 3]
    OUTPUT:
    P_6d  - the 6 DOF poses of size [batch_size, 6], where the first 3 elements of each row are the angle-axis rotation 
    vector (Euler vector) and the last 3 elements are the translation vector. 
    NOTE:
    For situations where all sets of 2D points in the mini-batch correspond to one common set of 3D points, use the BPnP class. 
    """
    @staticmethod
    def forward(ctx, pts2d, pts3d, K, ini_pose=None):
        bs = pts2d.size(0)
        n = pts2d.size(1)
        device = pts2d.device
        K_np = np.array(K.detach().cpu())
        P_6d = torch.zeros(bs,6,device=device)

        for i in range(bs):
            pts2d_i_np = np.ascontiguousarray(pts2d[i].detach().cpu()).reshape((n,1,2))
            pts3d_i_np = np.ascontiguousarray(pts3d[i].detach().cpu()).reshape((n,3))
            if ini_pose is None:
                _, rvec0, T0 = cv2.solvePnP(
                    objectPoints=pts3d_i_np, 
                    imagePoints=pts2d_i_np, 
                    cameraMatrix=K_np, 
                    distCoeffs=None,
                    flags=cv2.SOLVEPNP_ITERATIVE, 
                    useExtrinsicGuess=False)
            else:
                rvec0 = np.array(ini_pose[i, 0:3].cpu().reshape(3, 1))
                T0 = np.array(ini_pose[i, 3:6].cpu().reshape(3, 1))
            _, rvec, T = cv2.solvePnP(objectPoints=pts3d_i_np, imagePoints=pts2d_i_np, cameraMatrix=K_np, distCoeffs=None, flags=cv2.SOLVEPNP_ITERATIVE, useExtrinsicGuess=True, rvec=rvec0, tvec=T0)
            angle_axis = torch.tensor(rvec,device=device,dtype=torch.float).reshape(1, 3)
            T = torch.tensor(T,device=device,dtype=torch.float).reshape(1, 3)
            P_6d[i,:] = torch.cat((angle_axis,T),dim=-1)

        ctx.save_for_backward(pts2d,P_6d,pts3d,K)
        return P_6d

    @staticmethod
    def backward(ctx, grad_output):
        pts2d, P_6d, pts3d, K = ctx.saved_tensors
        device = pts2d.device
        bs = pts2d.size(0)
        n = pts2d.size(1)
        m = 6

        grad_x = torch.zeros_like(pts2d)
        grad_z = torch.zeros_like(pts3d)
        grad_K = torch.zeros_like(K)

        for i in range(bs):
            J_fy = torch.zeros(m,m, device=device)
            J_fx = torch.zeros(m,2*n, device=device)
            J_fz = torch.zeros(m,3*n, device=device)
            J_fK = torch.zeros(m, 9, device=device)

            torch.set_grad_enabled(True)
            pts2d_flat = pts2d[i].clone().reshape(-1).detach().requires_grad_()
            P_6d_flat = P_6d[i].clone().reshape(-1).detach().requires_grad_()
            pts3d_flat = pts3d[i].clone().reshape(-1).detach().requires_grad_()
            K_flat = K.clone().reshape(-1).detach().requires_grad_()

            for j in range(m):
                torch.set_grad_enabled(True)
                if j > 0:
                    pts2d_flat.grad.zero_()
                    P_6d_flat.grad.zero_()
                    pts3d_flat.grad.zero_()
                    K_flat.grad.zero_()

                R = kn.angle_axis_to_rotation_matrix(P_6d_flat[0:m-3].reshape(1,3))

                P = torch.cat((R[0,0:3,0:3].reshape(3,3), P_6d_flat[m-3:m].reshape(3,1)),dim=-1)
                KP = torch.mm(K_flat.reshape(3,3), P)
                pts2d_i = pts2d_flat.reshape(n,2).transpose(0,1)
                pts3d_i = torch.cat((pts3d_flat.reshape(n,3),torch.ones(n,1,device=device)),dim=-1).t()
                proj_i = KP.mm(pts3d_i)
                Si = proj_i[2,:].reshape(1,n)

                r = pts2d_i*Si-proj_i[0:2,:]
                coefs = _get_coefs(P_6d_flat.reshape(1,6), pts3d_flat.reshape(n,3), K_flat.reshape(3,3))
                coef = coefs[:,:,j].transpose(0,1) # size: [2,n]
                fj = (coef*r).sum()
                fj.backward()
                J_fy[j,:] = P_6d_flat.grad.clone()
                J_fx[j,:] = pts2d_flat.grad.clone()
                J_fz[j,:] = pts3d_flat.grad.clone()
                J_fK[j,:] = K_flat.grad.clone()

            inv_J_fy = torch.inverse(J_fy)

            J_yx = (-1) * torch.mm(inv_J_fy, J_fx)
            J_yz = (-1) * torch.mm(inv_J_fy, J_fz)
            J_yK = (-1) * torch.mm(inv_J_fy, J_fK)

            grad_x[i] = grad_output[i].reshape(1,m).mm(J_yx).reshape(n,2)
            grad_z[i] = grad_output[i].reshape(1,m).mm(J_yz).reshape(n,3)
            grad_K += grad_output[i].reshape(1,m).mm(J_yK).reshape(3,3)

        return grad_x, grad_z, grad_K, None



class BPnPFunction(torch.autograd.Function):
    r"""Back-propagatable PnP
        Arguments:
            pts2d - the 2D keypoints coordinates of size [batch_size, num_keypoints, 2]
            pts3d - the 3D keypoints coordinates of size [num_keypoints, 3]
            K     - the camera intrinsic matrix of size [3, 3]
        Returns:
            P_6d  - the 6 DOF poses of size [batch_size, 6], where the first 3 elements of each row are the angle-axis rotation 
            vector (Euler vector) and the last 3 elements are the translation vector. 
        NOTE: This bpnp function assumes that all sets of 2D points in the mini-batch correspond to one common set of 3D points. 
    """
    @staticmethod
    def forward(ctx, pts2d, pts3d, K, ini_pose=None):
        bs = pts2d.size(0)
        n = pts2d.size(1)
        device = pts2d.device
        pts3d_np = np.array(pts3d.detach().cpu())
        K_np = np.array(K.detach().cpu())
        P_6d = torch.zeros(bs,6,device=device)

        try:
            distCoeffs = pnp.distCoeffs
        except:
            distCoeffs = np.zeros((8, 1), dtype='float32') 

        for i in range(bs):
            pts2d_i_np = np.ascontiguousarray(pts2d[i].detach().cpu()).reshape((n, 1, 2))
            if ini_pose is None:
                _, rvec0, T0 = cv2.solvePnP(
                    objectPoints=pts3d_np, 
                    imagePoints=pts2d_i_np, 
                    cameraMatrix=K_np, 
                    distCoeffs=distCoeffs,
                    flags=cv2.SOLVEPNP_ITERATIVE, 
                    useExtrinsicGuess=False
                )
            else:
                rvec0 = np.array(ini_pose[0, 0:3].cpu().reshape(3, 1))
                T0 = np.array(ini_pose[0, 3:6].cpu().reshape(3, 1))
            _, rvec, T = cv2.solvePnP(
                objectPoints=pts3d_np,
                imagePoints=pts2d_i_np,
                cameraMatrix=K_np,
                distCoeffs=None,
                flags=cv2.SOLVEPNP_ITERATIVE, 
                useExtrinsicGuess=True, 
                rvec=rvec0,
                tvec=T0
            )
            angle_axis = torch.Tensor(rvec).reshape(1, 3).float().to(device)
            T = torch.Tensor(T).reshape(1, 3).float().to(device)
            P_6d[i, :] = torch.cat((angle_axis, T),dim=-1)

        ctx.save_for_backward(pts2d,P_6d,pts3d,K)
        return P_6d

    @staticmethod
    def backward(ctx, grad_output):
        pts2d, P_6d, pts3d, K = ctx.saved_tensors
        device = pts2d.device
        bs = pts2d.size(0)
        n = pts2d.size(1) #nof keypoints
        m = 6

        grad_x = torch.zeros_like(pts2d)
        grad_z = torch.zeros_like(pts3d)
        grad_K = torch.zeros_like(K)

        for i in range(bs):
            J_fy = torch.zeros(m, m, device=device)
            J_fx = torch.zeros(m, 2 * n, device=device)
            J_fz = torch.zeros(m, 3 * n, device=device)
            J_fK = torch.zeros(m, 9, device=device)

            torch.set_grad_enabled(True)
            pts2d_flat = pts2d[i].clone().reshape(-1).detach().requires_grad_()
            P_6d_flat = P_6d[i].clone().reshape(-1).detach().requires_grad_()
            pts3d_flat = pts3d.clone().reshape(-1).detach().requires_grad_()
            K_flat = K.clone().reshape(-1).detach().requires_grad_()

            for j in range(m):
                if j > 0:
                    pts2d_flat.grad.zero_()
                    P_6d_flat.grad.zero_()
                    pts3d_flat.grad.zero_()
                    K_flat.grad.zero_()

                R = kn.angle_axis_to_rotation_matrix(P_6d_flat[0: m-3].reshape(1, 3))

                P = torch.cat((
                    R[0, 0:3, 0:3].reshape(3, 3), P_6d_flat[m-3:m].reshape(3, 1)
                ), dim=-1)
                KP = torch.mm(K_flat.reshape(3, 3), P)
                pts2d_i = pts2d_flat.reshape(n, 2).transpose(0, 1)
                pts3d_i = torch.cat((
                    pts3d_flat.reshape(n, 3), torch.ones(n, 1, device=device)
                ), dim=-1).t()
                proj_i = KP.mm(pts3d_i)
                Si = proj_i[2, :].reshape(1, n)

                r = pts2d_i * Si - proj_i[0:2, :]
                
                coefs = _get_coefs(
                    P_6d_flat.reshape(1, 6),
                    pts3d_flat.reshape(n, 3),
                    K_flat.reshape(3, 3)
                )
                
                coef = coefs[:, :, j].transpose(0, 1) # size: [2,n]
                fj = (coef * r).sum()
                fj.backward()
                J_fy[j, :] = P_6d_flat.grad.clone()
                J_fx[j, :] = pts2d_flat.grad.clone()
                J_fz[j, :] = pts3d_flat.grad.clone()
                J_fK[j, :] = K_flat.grad.clone()

            inv_J_fy = torch.inverse(J_fy)

            J_yx = (-1.0) * torch.mm(inv_J_fy, J_fx)
            J_yz = (-1.0) * torch.mm(inv_J_fy, J_fz)
            J_yK = (-1.0) * torch.mm(inv_J_fy, J_fK)

            grad_x[i] = grad_output[i].reshape(1, m).mm(J_yx).reshape(n, 2)
            grad_z += grad_output[i].reshape(1, m).mm(J_yz).reshape(n, 3)
            grad_K += grad_output[i].reshape(1, m).mm(J_yK).reshape(3, 3)

        return grad_x, grad_z, grad_K, None

class BPnP(torch.nn.Module):

    def __init__(self,
        isBatch: bool=False, # whether 3D keypoints is a mini-batch
        transpose: bool=True, #whether to transpose the camera matrix
        useFast: bool=False, #whether to use faster implementation of BPnP 
    ):
        super(BPnP, self).__init__()

        self.BPnP_func = BPnPFunction if not isBatch else BPnPFunction_fast \
            if useFast else BPnPFunction_m3d
        self.transpose = transpose

    def forward(self, 
        keypoints2d:        torch.Tensor,       # [B, K, 2]
        intrinsics:         torch.Tensor,       # [3, 3]
        keypoints3d:        torch.Tensor,       # [K, 3] or # [B,K,3]
    ) -> torch.Tensor: # [B, 6]
            
        intrinsics = intrinsics[0,:,:] if len(intrinsics.shape) == 3 else intrinsics
        intrinsics = intrinsics.t() if self.transpose else intrinsics
        P_6d = self.BPnP_func.apply(keypoints2d, keypoints3d, intrinsics)
        
        return P_6d


from collections.abc import Callable

import torch
import typing
import omegaconf.omegaconf
import logging

log = logging.getLogger(__name__)

__all__ = ['NoOp']

class NoOp(Callable):
    def __init__(self, 
        calculators: omegaconf.DictConfig=None,
    ):
        log.info("No data statistics used.")

    def __call__(self, tensors: typing.Dict[str, torch.Tensor]) -> None:        
        pass
        


from moai.nn.activation.torch import (
    BN2d,
    ReLu_BN,
    BN2d_ReLu,
    ReLu_BN2d,
    LReLu_BN,
    LReLu_BN2d,
    BN2d_LReLu,
    Normalize,
)
from moai.nn.activation.snake import Snake
from moai.monads.math import Abs

import moai.nn.utils as miu

import torch
import logging

log = logging.getLogger(__name__)

__all__ = [
    "make_activation",
]

__ACTIVATION_FACTORY__ = {
    "bn2d":                     BN2d,
    "relu_bn":                  ReLu_BN,
    "bn2d_relu":                BN2d_ReLu,
    "relu_bn2d":                ReLu_BN2d,
    "lrelu_bn":                 LReLu_BN,
    "bn2d_lrelu":               BN2d_ReLu,
    "lrelu_bn2d":               LReLu_BN2d,
    "elu":                      torch.nn.ELU,
    "relu":                     torch.nn.ReLU,
    "abs":                      Abs,
    "none":                     torch.nn.Identity,
    "identity":                 torch.nn.Identity,
    "sigmoid":                  torch.nn.Sigmoid,
    "tanh":                     torch.nn.Tanh,
    "snake":                    Snake,
    "normalize":                Normalize,
}

def make_activation(
        activation_type: str,        
        features: int,                
        inplace: bool,
        **kwargs
    ) -> torch.nn.Module:
        if activation_type in __ACTIVATION_FACTORY__.keys():
            return miu.instantiate(__ACTIVATION_FACTORY__[activation_type], **{**locals(), **kwargs})
        else:
            log.error(f"Activation type ({activation_type}) not found.")



# Fixes a bug in original diffGrad code
import math
import torch
import numpy as np
import torch.nn as nn
import typing

__all__ = ['DiffGrad']

class DiffGrad(torch.optim.Optimizer):
    r"""Implements diffGrad algorithm. It is modified from the pytorch implementation of Adam.

    - **Paper**: [diffGrad: An Optimization Method for Convolutional Neural Networks](https://arxiv.org/pdf/1909.11015v3.pdf)
    - **Implementation**: [GitHub @ shivram1987](https://github.com/shivram1987/diffGrad)
    
    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
            (default: False)
    .. _diffGrad: An Optimization Method for Convolutional Neural Networks:
        https://arxiv.org/abs/1909.11015
    .. _Adam: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """

    def __init__(self, 
        params: typing.Iterator[torch.nn.Parameter],
        lr: float=1e-3,
        betas: typing.Tuple[float, float]=(0.9, 0.999),
        eps: float=1e-8,
        weight_decay: float=0
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        super(DiffGrad, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(DiffGrad, self).__setstate__(state)

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError('diffGrad does not support sparse gradients, please consider SparseAdam instead')

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p.data)
                    # Previous gradient
                    state['previous_grad'] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq, previous_grad = state['exp_avg'], state['exp_avg_sq'], state['previous_grad']
                beta1, beta2 = group['betas']

                state['step'] += 1

                if group['weight_decay'] != 0:
                    grad.add_(group['weight_decay'], p.data)

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(1 - beta1, grad)
                exp_avg_sq.mul_(beta2).addcmul_(1 - beta2, grad, grad)
                denom = exp_avg_sq.sqrt().add_(group['eps'])

                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                # compute diffgrad coefficient (dfc)
                diff = abs(previous_grad - grad)
                dfc = 1. / (1. + torch.exp(-diff))
                #state['previous_grad'] = grad %used in paper but has the bug that previous grad is overwritten with grad and diff becomes always zero. Fixed in the next line.
                state['previous_grad'] = grad.clone()
				
				# update momentum with dfc
                exp_avg1 = exp_avg * dfc

                step_size = group['lr'] * math.sqrt(bias_correction2) / bias_correction1

                p.data.addcdiv_(-step_size, exp_avg1, denom)

        return loss


import moai.nn.convolution.spherical as mis
import moai.nn.sampling.spatial.downsample.conv as midsc

import torch
import numpy as np
import logging

log = logging.getLogger(__name__)

__all__ = [
    "AntialiasedMaxPool2d",
    "AntialiasedStridedConv2d",
]

class AntialiasedMaxPool2d(torch.nn.MaxPool2d):
    def __init__(self,
        features: int,
        kernel_size: int=2,
        stride: int=2,
        padding: int=0,
        pad_type: str='reflect',
    ):
        super(AntialiasedMaxPool2d, self).__init__(
            kernel_size, stride=1, padding=int(np.ceil(1.0 * (kernel_size - 1.0) / 2.0))
        )
        self.filt_size = kernel_size
        self.pad_off = 0
        self.pad_sizes = [
            int(1.0 * (kernel_size - 1) / 2.0), int(np.ceil(1.0 * (kernel_size - 1.0) / 2.0)),
            int(1.0 * (kernel_size - 1) / 2.0), int(np.ceil(1.0 * (kernel_size - 1) / 2.0))
        ]
        self.pad_sizes = [pad_size + self.pad_off for pad_size in self.pad_sizes]
        self.aa_stride = stride if stride else kernel_size #NOTE: different name due to subclassing
        self.off = int((self.aa_stride - 1) / 2.0)
        self.channels = features

        if(self.filt_size==1):
            a = np.array([1.,])
        elif(self.filt_size==2):
            a = np.array([1., 1.])
        elif(self.filt_size==3):
            a = np.array([1., 2., 1.])
        elif(self.filt_size==4):    
            a = np.array([1., 3., 3., 1.])
        elif(self.filt_size==5):    
            a = np.array([1., 4., 6., 4., 1.])
        elif(self.filt_size==6):    
            a = np.array([1., 5., 10., 10., 5., 1.])
        elif(self.filt_size==7):    
            a = np.array([1., 6., 15., 20., 15., 6., 1.])

        filt = torch.Tensor(a[:,None] * a[None,:])
        filt = filt / torch.sum(filt)
        self.register_buffer('filter', filt[None,None,:,:].repeat((self.channels, 1, 1, 1)))
        self.pad = self._get_pad_layer(pad_type)(self.pad_sizes)

    def _get_pad_layer(self, pad_type: str):
        if(pad_type in ['refl', 'reflect']):
            PadLayer = torch.nn.ReflectionPad2d
        elif(pad_type in ['repl', 'replicate']):
            PadLayer = torch.nn.ReplicationPad2d
        elif(pad_type=='zero'):
            PadLayer = torch.nn.ZeroPad2d
        elif(pad_type=="spherical"):
            PadLayer = mis.SphericalPad2d
        else:
            log.error('Pad type [%s] not recognized' % pad_type)
        return PadLayer

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        pooled = super(AntialiasedMaxPool2d, self).forward(input)
        if(self.kernel_size==1):
            if(self.pad_off==0):
                return pooled[:,:,::self.aa_stride,::self.aa_stride]    
            else:
                return self.pad(pooled)[:,:,::self.aa_stride,::self.aa_stride]
        else:
            return torch.nn.functional.conv2d(self.pad(pooled), 
                self.filter, stride=self.aa_stride, groups=pooled.shape[1]
            )

class AntialiasedStridedConv2d(midsc.StridedConv2d):
    def __init__(self,
        features: int,
        kernel_size: int=3,
        stride: int=2,
        padding: int=1,
        pad_type: str='reflect',
        conv_type: str="conv2d",
    ):
        super(AntialiasedStridedConv2d, self).__init__(
            features=features, kernel_size=kernel_size,
            stride=1, padding=padding, conv_type=conv_type,
        )
        self.filt_size = kernel_size
        self.pad_off = 0
        self.pad_sizes = [
            int(1.0 * (kernel_size - 1) / 2.0), int(np.ceil(1.0 * (kernel_size - 1.0) / 2.0)),
            int(1.0 * (kernel_size - 1) / 2.0), int(np.ceil(1.0 * (kernel_size - 1) / 2.0))
        ]
        self.pad_sizes = [pad_size + self.pad_off for pad_size in self.pad_sizes]
        self.stride = stride
        self.off = int((self.stride - 1) / 2.0)
        self.channels = features

        if(self.filt_size==1):
            a = np.array([1.,])
        elif(self.filt_size==2):
            a = np.array([1., 1.])
        elif(self.filt_size==3):
            a = np.array([1., 2., 1.])
        elif(self.filt_size==4):    
            a = np.array([1., 3., 3., 1.])
        elif(self.filt_size==5):    
            a = np.array([1., 4., 6., 4., 1.])
        elif(self.filt_size==6):    
            a = np.array([1., 5., 10., 10., 5., 1.])
        elif(self.filt_size==7):    
            a = np.array([1., 6., 15., 20., 15., 6., 1.])

        filt = torch.Tensor(a[:,None]*a[None,:])
        filt = filt / torch.sum(filt)
        self.register_buffer('filter', filt[None,None,:,:].repeat((self.channels, 1, 1, 1)))
        self.pad = self._get_pad_layer(pad_type)(self.pad_sizes)

    def _get_pad_layer(self, pad_type: str):
        if(pad_type in ['refl', 'reflect']):
            PadLayer = torch.nn.ReflectionPad2d
        elif(pad_type in ['repl', 'replicate']):
            PadLayer = torch.nn.ReplicationPad2d
        elif(pad_type=='zero'):
            PadLayer = torch.nn.ZeroPad2d
        elif(pad_type=="spherical"):
            PadLayer = mis.SphericalPad2d
        else:
            log.error('Pad type [%s] not recognized' % pad_type)
        return PadLayer

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        pooled = super(AntialiasedStridedConv2d, self).forward(input)
        if(self.kernel_size==1):
            if(self.pad_off==0):
                return pooled[:,:,::self.stride,::self.stride]    
            else:
                return self.pad(pooled)[:,:,::self.stride,::self.stride]
        else:
            return torch.nn.functional.conv2d(self.pad(pooled), 
                self.filter, stride=self.stride, groups=pooled.shape[1]
            )


import torch

__all__ = ["Binary"]

class Binary(torch.nn.Module):
    def __init__(self):
        super(Binary, self).__init__()

    def forward(self, tensor: torch.Tensor) -> torch.Tensor:
        return (tensor != 0.0).bool() #TODO: byte or bool?


from collections import namedtuple
from turtle import width
import torch
import typing
import logging
import numpy as np

__all__ = ['Silhouette']

log = logging.getLogger(__name__)

try:
    import pyredner
    from moai.monads.render.redner.render import render_g_buffer
except:
    log.error(f"The pyredner package (https://github.com/BachiLi/redner) is required to use the corresponding rendering monads. Install with `pip install pyredner`")

RenderParams = namedtuple('RenderParams', [
    'samples',
    'sample_pixel_center',
    'focal_length',
    'near_clip',
    'resolution',
    'principal_point',
    'opengl',
])

class Silhouette(torch.nn.Module):    
    def __init__(self,
        resolution:             typing.Sequence[int]=(256, 512),
        near_clip:              float=0.001,        
        samples:                typing.Union[int, typing.Sequence[int]]=(16, 4),
        sample_pixel_center:    bool=False,
        focal_length:           typing.Union[typing.Sequence[float], float]=5000.0,
        principal_point:        typing.Union[typing.Sequence[float], float]=None,
        opengl:                 bool=True,
    ):
        super(Silhouette, self).__init__()
        self.params = RenderParams(
            samples=(samples, samples) if isinstance(samples, int) else samples,
            resolution=resolution, opengl=opengl,
            sample_pixel_center=sample_pixel_center, near_clip=near_clip,
            focal_length=(float(focal_length), float(focal_length)) if isinstance(focal_length, (float, int)) else focal_length, 
            principal_point=(float(principal_point), float(principal_point)) if isinstance(principal_point, (float, int)) else principal_point
        )

    def forward(self,
        vertices:           torch.Tensor,
        indices:            torch.Tensor,
        uvs:                torch.Tensor=None,
        camera_translation: torch.Tensor=None,
        camera_rotation:    torch.Tensor=None,
        image:              torch.Tensor=None,
        intrinsics:         torch.Tensor=None,
    ) -> typing.Dict[str, torch.Tensor]:
        b, n, _ = vertices.shape
        translation = camera_translation if camera_translation is not None\
            else torch.zeros(b, 3, device=vertices.device)
        rotation = camera_rotation if camera_rotation is not None\
            else torch.eye(3, device=vertices.device)[np.newaxis, ...].expand(b, 3, 3)
        resolution = self.params.resolution or image.shape[-2:]
        if intrinsics is None:
            with torch.no_grad():
                ndc_width = 2.0 * resolution[-1]
                width_scale = resolution[-1] / image.shape[-1]
                height_scale = resolution[-2] / image.shape[-2]
                intrinsics = torch.eye(3, device=vertices.device)[np.newaxis, ...].expand(b, 3, 3).clone()
                intrinsics[:, 0, 0] = (self.params.focal_length[0] / ndc_width) * width_scale
                intrinsics[:, 1, 1] = (self.params.focal_length[1] / ndc_width) * height_scale
                intrinsics[:, 0, 0] *= 4.0
                intrinsics[:, 1, 1] *= 4.0
                # intrinsics[:, 0, 2] = (
                #     (self.params.principal_point[0] - resolution[-1] / 2)
                #     /  (8.0 * resolution[-1])
                #     # /  resolution[-1] / 2
                # ) if self.params.principal_point is not None else 0.0 # resolution[-1] / 2
                # intrinsics[:, 1, 2] = (
                #     (self.params.principal_point[1] - resolution[-2] / 2)
                #     # / resolution[-2] / 2
                #     / (8.0 * resolution[-2])
                # ) if self.params.principal_point is not None else 0.0 # resolution[-2] / 2
                intrinsics[:, 0, 2] = (
                    (0.5 * resolution[-1] - self.params.principal_point[0] * width_scale)
                    / (0.5 * resolution[-1]) / 1.0
                ) if self.params.principal_point is not None else 0.0
                #NOTE: should be 1/aspect_ratio, thus normalized with width resolution
                #NOTE: see https://github.com/BachiLi/redner/issues/83 
                intrinsics[:, 1, 2] = (
                    (0.5 * resolution[-2] - self.params.principal_point[1] * height_scale)
                    / (0.5 * resolution[-1]) / 1.0
                ) if self.params.principal_point is not None else 0.0
        # if self.params.resolution is not None and image is not None:
        #     scale_y = image.shape[-2] / self.params.resolution[0]
        #     scale_x = image.shape[-1] / self.params.resolution[1]
        #     intrinsics[:, 0, :] *= scale_x
        #     intrinsics[:, 1, :] *= scale_y
        kwargs = { 
            'material': pyredner.Material(
                diffuse_reflectance=torch.ones_like(translation[0]),
                use_vertex_color=True,
            )
        }
        scenes = []
        for i in range(b):
            f = indices[i].int().contiguous()
            v = vertices[i].contiguous()
            # v[..., 0] = -1.0 * v[..., 0] # v3
            # v[..., 1] = -1.0 * v[..., 1] # v3
            # v[..., 2] = -1.0 * v[..., 2] # v3
            v_r = -v
            if uvs is not None:
                kwargs['uvs'] = uvs[i]
            else:
                kwargs['colors'] = torch.ones_like(v)    
            obj = pyredner.Object(vertices=v_r, indices=f,**kwargs)
            pos = translation[i]
            # if self.params.opengl:
                # # pos = pos * torch.tensor([-1.0, -1.0, 1.0]).to(pos)
                # pos = pos * torch.tensor([-1.0, 1.0, 1.0]).to(pos) # v3
            # up = (-1.0 if self.params.opengl else 1.0) * rotation[i][1]
            up = rotation[i][1] # v3
            fwd = (-1.0 if self.params.opengl else 1.0) * rotation[i][2]
            cam = pyredner.Camera(
                position=pos, up=up, look_at=pos + fwd,
                clip_near=self.params.near_clip,
                resolution=resolution,
                intrinsic_mat=intrinsics[i], 
                camera_type=pyredner.camera_type.perspective
            )
            scene = pyredner.Scene(camera=cam, objects=[obj])
            scenes.append(scene)
        img = render_g_buffer(scene=scenes, 
            channels=[pyredner.channels.vertex_color], seed=None,                           
            num_samples=(self.params.samples[0], self.params.samples[1]),            
            sample_pixel_center=self.params.sample_pixel_center,
            device=vertices.device
        )
        img = img.permute(0, 3, 1, 2).contiguous()
        img = img.mean(dim=1, keepdim=True)
        if self.params.opengl:
            img = img.flip(dims=[-1])
        return torch.clamp(img, min=0.0, max=1.0)


from moai.validation.metrics.object_pose.position import NormalizedPositionError
from moai.validation.metrics.object_pose.rotation import AngleError
from moai.validation.metrics.object_pose.accuracy import (
    Accuracy2,
    Accuracy5,
    Accuracy10
)
from moai.validation.metrics.object_pose.add import (
    SixD2,
    SixD5,
    SixD10
)

from moai.validation.metrics.object_pose.projection import (
    Projection2,
    Projection5,
    Projection10
)

__all__ = [
    "NormalizedPositionError",
    "AngleError",
    "Accuracy2",
    "Accuracy5",
    "Accuracy10",
    "SixD2",
    "SixD5",
    "SixD10",
    "Projection",
    "Projection",
    "Projection10",
]


import moai.nn.deconvolution as mid

import torch

__all__ = [
    "StridedDeconv2d",
]

class StridedDeconv2d(torch.nn.Module):
    def __init__(self,
        features: int,
        kernel_size: int=4,
        deconv_type: str="deconv2d",
        stride: int=2,
        padding: int=1
    ):
        super(StridedDeconv2d, self).__init__()
        self.deconv = mid.make_deconv_op(
            deconvolution_type=deconv_type,
            in_channels=features,
            out_channels=features,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.deconv(x)


import moai.parameters.optimization.optimizers as mio
from moai.parameters.optimization.optimizers.lookahead import Lookahead

import torch
import math
import torch
import typing

__all__ = ['RangerLars']

#NOTE: from https://github.com/mgrankin/over9000/blob/master/rangerlars.py

class RangerLars(Lookahead):
     """Implements a combination of RAdam, LARS and Lookahead
     
     - [ON THE VARIANCE OF THE ADAPTIVE LEARNING RATE AND BEYOND](https://arxiv.org/pdf/1908.03265.pdf)
     - [Lookahead Optimizer: k steps forward, 1 step back](https://arxiv.org/pdf/1907.08610.pdf)
     - [Large Batch Training of Convolutional Networks](https://arxiv.org/abs/1708.03888)
     - **Implementation**: [GitHub @ mgrankin](https://github.com/mgrankin/over9000)

     """
     def __init__(self,
          params: typing.Iterator[torch.nn.Parameter],
          alpha: float=0.5,
          k: int=5,
          lr: float=1e-3,
          betas: typing.Tuple[float, float]=(0.9, 0.999),
          eps: float=1e-8,
          weight_decay: float=0
     ):
          super(RangerLars, self).__init__(
               mio.Ralamb(params,
                    lr=lr, betas=betas, eps=eps, weight_decay=weight_decay,
               ), k=k, alpha=alpha
          )


import hydra
import sys
import os
import omegaconf.omegaconf
import logging
import typing
import ipaddress
from rich.console import Console

log = logging.getLogger(__name__)

def assign(cfg: omegaconf.DictConfig, attr: str) -> typing.Union[typing.Any]:
    return getattr(cfg, attr) if hasattr(cfg, attr) else None

def isIP(ip_address: str):
    ip_address = ip_address.replace('localhost', '127.0.0.1')
    try:
        ipaddress.ip_address(ip_address)
        return True
    except:
        pass

def search_obj(obj: omegaconf.DictConfig, console: Console, var_path=[]):
    for k, v in obj.items():
        var_path.append(k)
        if not isinstance(v, omegaconf.DictConfig):
            if isinstance(v, str):
                values = os.path.splitext(v)
                if k == 'gpus':
                    full_key = '.'.join(var_path)
                    value = ''
                    while not (
                        (len(value) == 1 and value.isdigit()) or \
                        (len(value) == 2 and value[1:].isdigit() and int(value) > -2 and int(value) < 10)) \
                        and not value == 's': # NOTE: to check the conditions for GPUs values
                        value = console.input(f"==> {full_key} ([italic cyan]{obj[k]}[/])? [[bold yellow]s (skip)[/]] ").lower()
                    if not value=='s':
                        obj[k] = value
                        console.print(f"[green]Param set[/] --> ([italic cyan]{obj[k]}[/])")
                    else:
                        console.print(f"[yellow]Param skipped[/] --> ([italic cyan]{obj[k]}[/])")
                elif len(os.path.dirname(values[0])):
                    full_key = '.'.join(var_path)
                    value = ''
                    while not len(os.path.dirname(value)) and not value == 's':
                        value = console.input(f"==> {full_key} ([italic cyan]{obj[k]}[/])? [[bold yellow]s (skip)[/]] ").lower()
                    if not value=='s':
                        obj[k] = value
                        console.print(f"[green]Param set[/] --> ([italic cyan]{obj[k]}[/])")
                    else:
                        console.print(f"[yellow]Param skipped[/] --> ([italic cyan]{obj[k]}[/])")
                else:
                    if isIP(v):
                        full_key = '.'.join(var_path)
                        value = ''
                        while not (isinstance(value, str) and isIP(value)) and not value == 's':
                            value = console.input(f"==> {full_key} ([italic cyan]{obj[k]}[/])? [[bold yellow]s (skip)[/]] ").lower()
                        if not value=='s':
                            obj[k] = value
                            console.print(f"[green]Param set[/] --> ([italic cyan]{obj[k]}[/])")
                        else:
                            console.print(f"[yellow]Param skipped[/] --> ([italic cyan]{obj[k]}[/])")
            var_path.pop(-1)
        else:
            search_obj(v, console, var_path)
    if len(var_path):
        var_path.pop(-1)

def env_query(cfg: omegaconf.DictConfig, console: Console):
    command = console.input("[bold yellow]Is the environment the one where the experiment has been executed? (y/n) [/]").lower()
    if command == "y":
        pass
    elif command == "n":
        search_obj(cfg, console)
    else:
        env_query(cfg, console)

def reprod(cfg):    
    console = Console()
    console.rule("Environment Settings", style="bold blue")
    with omegaconf.open_dict(cfg):
        del cfg["reprod"]
        env_query(cfg, console)
    console.rule("Reproduction Begin", style="bold blue")
    hydra.utils.log.debug(f"Configuration:\n{omegaconf.OmegaConf.to_yaml(cfg, resolve=True)}")
    with open("config_resolved.yaml", 'w') as f:
        f.write(omegaconf.OmegaConf.to_yaml(cfg, resolve=True, sort_keys=False))
    with open("config.yaml", 'w') as f:
        f.write(omegaconf.OmegaConf.to_yaml(cfg, resolve=False, sort_keys=False))
    engine = hydra.utils.instantiate(cfg.engine)
    model = hydra.utils.instantiate(cfg.model,
        data=cfg.data,
        visualization=assign(cfg, "visualization"),
        export=assign(cfg, "export"),    
    )
    model.initialize_parameters()
    trainer = hydra.utils.instantiate(cfg.trainer, 
        logging=assign(cfg, "logging")
    )
    log.info("Training started.")     
    trainer.run(model)
    log.info("Training completed.")

if __name__ == "__main__":
    # os.environ['HYDRA_FULL_ERROR'] = '1'
    config_filename = get_conf_path(sys.argv.pop(1)) #TODO: argparser integration?
    sys.argv.append("hydra.run.dir=actions/${hydra.job.name}/${now:%Y-%m-%d}/${now:%H-%M-%S}-${experiment.name}")    
    reprod = hydra.main(config_path="conf", config_name=config_filename)(reprod)
    reprod()


try:
    from moai.action.train import train
    from moai.action.evaluate import evaluate
    from moai.action.play import play
    from moai.action.diff import diff
    from moai.action.plot import plot
    from moai.action.reprod import reprod
    from moai.action.fit import fit
    from moai.action.archive import archive
except:
    from action.train import train
    from action.evaluate import evaluate
    from action.play import play
    from action.diff import diff
    from action.plot import plot
    from action.reprod import reprod
    from action.fit import fit
    from action.archive import archive

import omegaconf.omegaconf
import hydra
import logging
import sys
import os

from rich.traceback import install
install(width=120, extra_lines=5, theme=None,
    word_wrap=True, show_locals=False, indent_guides=True,    
)

ERROR_FORMAT = "%(levelname)s at %(asctime)s in %(funcName)s in %(filename) at line %(lineno)d: %(message)s"
DEBUG_FORMAT = "%(lineno)d in %(filename)s at %(asctime)s: %(message)s"
# FORMAT = "[%(asctime)s][%(filename)s][%(levelname)s] - %(message)s"
FORMAT = "[%(levelname)s] - %(message)s"

logging.basicConfig(level=logging.INFO, format=FORMAT)
log = logging.getLogger('moai')

def debug(cfg):
    log.info(cfg)

__MODES__ = {
    'train': train,
    'evaluate': evaluate,
    'play': play,
    'diff': diff,
    'plot': plot,
    'debug': debug,
    'reprod': reprod,
    'fit': fit,
    'archive': archive,
}

__MIN_ARGS_COUNT__ = {
    'train': 2,
    'evaluate': 2,
    'play': 2,
    'diff': 1,
    'plot': 1,
    'debug': 2,
    'reprod': 2,
    'fit': 2,
    'archive': 2,
}

def run(cfg: omegaconf.DictConfig):
    reprod_key = "reprod"
    if not reprod_key in cfg:
        __MODES__[cfg.mode](cfg)
    else:
        __MODES__[cfg.reprod](cfg)

def moai():    
    # os.environ['HYDRA_FULL_ERROR'] = '1'
    mode = sys.argv.pop(1)
    if mode not in __MODES__:
        log.error(f"Wrong moai execution mode ({mode}), supported modes are: {list(__MODES__.keys())}.")
        sys.exit(-1)
    min_args = __MIN_ARGS_COUNT__[mode]
    if len(sys.argv) < min_args:
        log.error(f"Insufficient arguments provided for moai. Calling should specify the mode and config file: \'moai MODE CONFIG\'.")    
    config = sys.argv.pop(1) if min_args > 1 else f"tools/{mode}.yaml" 
    other_args = []
    if min_args > 1 or mode == 'plot':
        output_dir = "hydra.run.dir=actions/" + mode + "/${now:%Y-%m-%d}/${now:%H-%M-%S}-${experiment.name}"
        other_args.append(output_dir)
    else:
        other_args.append("+hydra.hydra_logging.root.handles=[]")
        other_args.append("hydra.hydra_logging.disable_existing_loggers=true")
        other_args.append("hydra.output_subdir=null")
    for oarg in other_args:
        sys.argv.append(oarg)
    if mode != 'reprod':
        sys.argv.append(f"+mode={mode}")
    else:
        sys.argv.append(f"+reprod={mode}")
    main = hydra.main(config_path="conf", config_name=config)(run)
    main()

if __name__ == "__main__":
    moai()


import torch
import kornia

__all__ = ["StructuralDisimilarity"]

#NOTE: check if kornia fixes its implementation https://github.com/kornia/kornia/issues/473 

class StructuralDisimilarity(kornia.losses.SSIM):
    def __init__(self,
        window_size: int=7,        
        dynamic_range: float=1.0
    ):
        super(StructuralDisimilarity, self).__init__(
            window_size=window_size,
            reduction='none',
            max_val=dynamic_range
        )

    def forward(self, 
        pred: torch.Tensor,
        gt: torch.Tensor,        
        weights: torch.Tensor=None,
        mask: torch.Tensor=None,
    ) -> torch.Tensor:                    
        if mask is not None:
            gt = torch.where(mask, gt, torch.zeros_like(gt))
            pred = torch.where(mask, pred, torch.zeros_like(gt))            
        ssim = super(StructuralDisimilarity, self).forward(gt, pred)
        if weights is not None:
            ssim = ssim * weights
        if mask is not None:
            ssim = ssim[mask]
        return torch.clamp(1.0 - (ssim + 1.0) * 0.5, min=0.0, max=1.0)


import moai.nn.residual as mires
import moai.nn.sampling.spatial.downsample as mids
import moai.nn.convolution as mic
import moai.nn.utils as miu

import torch
import itertools
import typing
import logging
import omegaconf.omegaconf

log = logging.getLogger(__name__)

#NOTE: https://towardsdatascience.com/an-overview-of-resnet-and-its-variants-5281e2f56035 
#NOTE: https://shuzhanfan.github.io/2018/11/ResNet/
#NOTE: https://towardsdatascience.com/illustrated-10-cnn-architectures-95d78ace614d

__all__ = ["ResNet"]

class ResNet(torch.nn.Module):
    def __init__(self,
        configuration:  omegaconf.DictConfig, # in & out features, input
        preproc:        omegaconf.DictConfig, # blocks, type, convolution, activation
        residual:       omegaconf.DictConfig, # blocks, type, convolution, activation
        downscale:      omegaconf.DictConfig, # 
    ):
        super(ResNet, self).__init__()
        self.block_modules = torch.nn.ModuleDict() # build network
        in_features, out_features = configuration.in_features, configuration.start_features            
        preproc_blocks = torch.nn.ModuleList()
        for i in range(0, preproc.block.count): # preproc blocks
            preproc_blocks.append(mic.make_conv_block(
                block_type=preproc.block.type,
                convolution_type=preproc.convolution.type,
                in_features=in_features,
                out_features=out_features,
                activation_type=preproc.activation.type,                    
                activation_params=preproc.activation.params,
                convolution_params={ 
                    k: miu.repeat(preproc.block.count, v)[i] for k, v in preproc.convolution.params.items() 
                }
            ))
            in_features = out_features
            out_features *= 2
        self.block_modules.add_module('preproc', torch.nn.Sequential(*preproc_blocks))
        block_names = ["block" + str(i) for i in range(1, len(residual.block.setup))]
        block_names.append("features")
        out_features = int(in_features * residual.output_factor)
        bottleneck_features = int(in_features * residual.bottleneck_factor)
        for i, (block_name, block_size) in enumerate(zip(block_names, residual.block.setup)): # residual blocks
            residual_blocks = torch.nn.ModuleList()
            residual_blocks.append(mids.make_downsample( # 1 downscale operation
                downscale_type=downscale.type,
                features=in_features,
                **downscale.params
            ))
            for b in range(0, block_size):
                residual_blocks.append(mires.make_residual_block( # b residual connections
                    block_type=residual.block.type,
                    convolution_type=residual.convolution.type,
                    in_features=(in_features if b == 0 else out_features),
                    out_features=out_features,
                    bottleneck_features=bottleneck_features,
                    activation_type=residual.activation.type,
                    strided=(b == 0 and downscale.type == "in_block"),
                    convolution_params=residual.convolution.params,
                    activation_params=residual.activation.params,
                    downscale_params=downscale.params
                ))
            self.block_modules.add_module(block_name, torch.nn.Sequential(*residual_blocks))                
            in_features = out_features
            bottleneck_features *= 2
            out_features *= 2

    def forward(self, 
        data: torch.Tensor 
    ) -> typing.Tuple[torch.Tensor, typing.List[torch.Tensor]]:
        features = [data]
        for _, block in self.block_modules.items():
            features.append(block(features[-1]))
        return features[-1], features[2:-1] #TODO: return stem features as well for UNet?


import torch

import functools

__all__ = [
    "VisibilityFOV",
    "VisibilityHeatmap"
]

class VisibilityFOV(torch.nn.Module):
    def __init__(self,
        width:          int=1,
        height:         int=1,
        coord_type:     str='coord',              # [B, K, UV]
    ):
        super(VisibilityFOV,self).__init__()
        self.coord_type = coord_type
        self.width = width
        self.height = height

    def forward(self,
        coords:   torch.Tensor
    ) -> torch.Tensor:  
        _coords = coords.clone().detach()
        if self.coord_type != 'coord':
            _coords[..., 0] = (_coords[..., 0] + 1.0) / 2.0 * self.width if self.coord_type == 'ndc' else _coords[..., 0] * self.width
            _coords[..., 1] = (_coords[..., 1] + 1.0) / 2.0 * self.height if self.coord_type == 'ndc' else _coords[..., 1] * self.height
        masks = torch.zeros_like(coords)
        masks[..., 0] = (_coords[..., 0] >= 0) * (_coords[..., 0] < self.width)
        masks[..., 1] = (_coords[..., 1] >= 0) * (_coords[..., 1] < self.height)
        return masks

class VisibilityHeatmap(torch.nn.Module):

    def _mask(self,
        coords:     torch.Tensor,           # [B, K, UV]
        heatmaps:   torch.Tensor            # [B, K, (D), H, W], with its value across the spatial dimensions summing to unity
    ) -> torch.Tensor: 
        r"""Extracts the visibility mask of keypoint predictions based on heatmap values.  .

        Args:
            coords:     torch.Tensor,               [B, K, UV(S) or (S)VU]
            heatmaps:   torch.Tensor                [B, K, (D), H, W], with its value across the spatial dimensions summing to unity

        Returns:
            The visibility weights of coordinates,  [B, K, UV(S) or (S)VU].
        """
        masks = torch.zeros_like(coords)
        channels = heatmaps.shape[1]
        for i in range(channels):
            heatmap = heatmaps[:, i, ...]
            for b in range(coords.shape[0]):
                uv = tuple(coords.flip(-1).long()[b, i])
                if uv[0] > -1 and uv[1] > -1 and uv[0] < heatmap.shape[-2] and uv[1] < heatmap.shape[-1]:
                    masks[b, i, ...] = 1.0 if heatmap[b][uv] > self.threshold else 0.0
                else:
                    masks[b, i, ...] = 0.0
        return masks

    def __init__(self,
        width:          int=1,
        height:         int=1,
        threshold:      float=0.4,
        coord_type:     str='coord',              # [B, K, UV]
    ):
        super(VisibilityHeatmap,self).__init__()
        self.coord_type = coord_type
        self.width = width
        self.height = height
        self.threshold = threshold

    def forward(self,
        coords:     torch.Tensor,           # [B, K, UV(S) or (S)VU]
        heatmaps:   torch.Tensor            # [B, K, (D), H, W], with its value across the spatial dimensions summing to unity
    ) -> torch.Tensor:                      # [B, K]
        _coords = coords.clone().detach()
        if self.coord_type != 'coord':
            _coords[..., 0] = (_coords[..., 0] + 1.0) / 2.0 * self.width if self.coord_type == 'ndc' else _coords[..., 0] * self.width
            _coords[..., 1] = (_coords[..., 1] + 1.0) / 2.0 * self.height if self.coord_type == 'ndc' else _coords[..., 1] * self.height
        masks = self._mask(_coords, heatmaps)
        return masks





# modified from https://github.com/mpyrozhok/adamwr

#NOTE: this is currently not supported

import torch
import torch.optim as pto
import torch.optim.lr_scheduler as ptlrs
import math
import typing
import logging

log = logging.getLogger(__name__)

class ReduceMaxLROnRestart:
    def __init__(self, ratio: float=0.75):
        self.ratio = ratio

    def __call__(self, eta_min: int, eta_max: int) -> float:
        return eta_min, eta_max * self.ratio

class ExpReduceMaxLROnIteration:
    def __init__(self, gamma: float=1.0):
        self.gamma = gamma

    def __call__(self, eta_min: int, eta_max: int, iterations: int) -> float:
        return eta_min, eta_max * self.gamma ** iterations

class CosinePolicy:
    def __call__(self, t_cur: int, restart_period: int) -> float:
        return 0.5 * (1. + math.cos(math.pi * (t_cur / restart_period)))

class ArccosinePolicy:
    def __call__(self, t_cur: int, restart_period: int):
        return (math.acos(max(-1, min(1, 2 * t_cur / restart_period - 1))) / math.pi)

class TriangularPolicy:
    def __init__(self, triangular_step: float=0.5):
        self.triangular_step = triangular_step

    def __call__(self, t_cur: int, restart_period: int):
        inflection_point = self.triangular_step * restart_period
        point_of_triangle = (t_cur / inflection_point
                             if t_cur < inflection_point
                             else 1.0 - (t_cur - inflection_point)
                             / (restart_period - inflection_point))
        return point_of_triangle

class CyclicLRWithRestarts(ptlrs._LRScheduler):
    """Decays learning rate with cosine annealing, normalizes weight decay
    hyperparameter value, implements restarts.
    https://arxiv.org/abs/1711.05101

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        batch_size: minibatch size
        epoch_size: training samples per epoch
        restart_period: epoch count in the first restart period
        t_mult: multiplication factor by which the next restart period will expand/shrink
        policy: ["cosine", "arccosine", "triangular", "triangular2", "exp_range"]
        min_lr: minimum allowed learning rate
        verbose: print a message on every restart
        gamma: exponent used in "exp_range" policy
        eta_on_restart_cb: callback executed on every restart, adjusts max or min lr
        eta_on_iteration_cb: callback executed on every iteration, adjusts max or min lr
        triangular_step: adjusts ratio of increasing/decreasing phases for triangular policy


    Example:
        >>> scheduler = CyclicLRWithRestarts(optimizer, 32, 1024, restart_period=5, t_mult=1.2)
        >>> for epoch in range(100):
        >>>     scheduler.step()
        >>>     train(...)
        >>>         ...
        >>>         optimizer.zero_grad()
        >>>         loss.backward()
        >>>         optimizer.step()
        >>>         scheduler.batch_step()
        >>>     validate(...)
    """

    def __init__(self, 
        optimizer: pto.Optimizer,
        batch_size: int,
        epoch_size: int,
        restart_period: int=100,
        t_mult: int=2, 
        last_epoch: int=-1,
        verbose: bool=False,
        policy: str="cosine", 
        policy_fn: typing.Callable=None,
        min_lr: float=1e-7,
        eta_on_restart_cb: typing.Callable=None,
        eta_on_iteration_cb: typing.Callable=None,
        gamma: float=1.0,
        triangular_step: float=0.5
    ):
        if not isinstance(optimizer, pto.Optimizer):
            raise TypeError('{} is not an Optimizer'.format(
                type(optimizer).__name__))
        self.optimizer = optimizer

        if last_epoch == -1:
            for group in optimizer.param_groups:
                group.setdefault('initial_lr', group['lr'])
                group.setdefault('minimum_lr', min_lr)
        else:
            for i, group in enumerate(optimizer.param_groups):
                if 'initial_lr' not in group:
                    raise KeyError("param 'initial_lr' is not specified "
                                   "in param_groups[{}] when resuming an"
                                   " optimizer".format(i))

        self.base_lrs = [group['initial_lr'] for group
                         in optimizer.param_groups]

        self.min_lrs = [group['minimum_lr'] for group
                        in optimizer.param_groups]

        self.base_weight_decays = [group['weight_decay'] for group
                                   in optimizer.param_groups]

        self.policy = policy
        self.eta_on_restart_cb = eta_on_restart_cb
        self.eta_on_iteration_cb = eta_on_iteration_cb
        if policy_fn is not None:
            self.policy_fn = policy_fn
        elif self.policy == "cosine":
            self.policy_fn = CosinePolicy()
        elif self.policy == "arccosine":
            self.policy_fn = ArccosinePolicy()
        elif self.policy == "triangular":
            self.policy_fn = TriangularPolicy(triangular_step=triangular_step)
        elif self.policy == "triangular2":
            self.policy_fn = TriangularPolicy(triangular_step=triangular_step)
            self.eta_on_restart_cb = ReduceMaxLROnRestart(ratio=0.5)
        elif self.policy == "exp_range":
            self.policy_fn = TriangularPolicy(triangular_step=triangular_step)
            self.eta_on_iteration_cb = ExpReduceMaxLROnIteration(gamma=gamma)

        self.last_epoch = last_epoch
        self.batch_size = batch_size
        self.epoch_size = epoch_size

        self.iteration = 0
        self.total_iterations = 0

        self.t_mult = t_mult
        self.verbose = verbose
        self.restart_period = math.ceil(restart_period)
        self.restarts = 0
        self.t_epoch = -1
        self.epoch = -1

        self.eta_min = 0
        self.eta_max = 1

        self.end_of_period = False
        self.batch_increments = []
        self._set_batch_increment()

    def _on_restart(self):
        if self.eta_on_restart_cb is not None:
            self.eta_min, self.eta_max = self.eta_on_restart_cb(self.eta_min,
                                                                self.eta_max)

    def _on_iteration(self):
        if self.eta_on_iteration_cb is not None:
            self.eta_min, self.eta_max = self.eta_on_iteration_cb(self.eta_min,
                                                                  self.eta_max,
                                                                  self.total_iterations)

    def get_lr(self, t_cur):
        eta_t = (self.eta_min + (self.eta_max - self.eta_min)
                 * self.policy_fn(t_cur, self.restart_period))

        weight_decay_norm_multi = math.sqrt(self.batch_size /
                                            (self.epoch_size *
                                             self.restart_period))

        lrs = [min_lr + (base_lr - min_lr) * eta_t for base_lr, min_lr
               in zip(self.base_lrs, self.min_lrs)]
        weight_decays = [base_weight_decay * eta_t * weight_decay_norm_multi
                         for base_weight_decay in self.base_weight_decays]

        if (self.t_epoch + 1) % self.restart_period < self.t_epoch:
            self.end_of_period = True

        if self.t_epoch % self.restart_period < self.t_epoch:
            if self.verbose:
                log.info(f"Restart {self.restarts + 1} at epoch {self.last_epoch}")
            self.restart_period = math.ceil(self.restart_period * self.t_mult)
            self.restarts += 1
            self.t_epoch = 0
            self._on_restart()
            self.end_of_period = False

        return zip(lrs, weight_decays)

    def _set_batch_increment(self) -> None:
        d, r = divmod(self.epoch_size, self.batch_size)
        batches_in_epoch = d + 2 if r > 0 else d + 1
        self.iteration = 0
        self.batch_increments = torch.linspace(0, 1, batches_in_epoch).tolist()

    def step(self) -> None:
        self.last_epoch += 1
        self.t_epoch += 1
        self._set_batch_increment()
        self.batch_step()

    def batch_step(self) -> None:
        try:
            t_cur = self.t_epoch + self.batch_increments[self.iteration]
            self._on_iteration()
            self.iteration += 1
            self.total_iterations += 1
        except (IndexError):
            raise StopIteration("Epoch size and batch size used in the "
                                "training loop and while initializing "
                                "scheduler should be the same.")

        for param_group, (lr, weight_decay) in zip(self.optimizer.param_groups,
                                                   self.get_lr(t_cur)):
            param_group['lr'] = lr
            param_group['weight_decay'] = weight_decay



import moai.nn.convolution as mic
import moai.nn.activation as mia

import torch

__all__ = [
    "Bottleneck",
    "PreResBottleneck",
    "PreActivBottleneck",
]

'''
    Bottleneck versions with 3 convolutions (2 projections, 1 bottleneck)    
'''
class Bottleneck(torch.nn.Module):
    def __init__(self,
        convolution_type: str,
        activation_type: str,
        in_features: int,
        out_features: int,
        bottleneck_features: int,
        convolution_params: dict,
        activation_params: dict,
        strided: bool,
    ):
        super(Bottleneck, self).__init__()
        self.W1 = mic.make_conv_1x1(
            convolution_type=convolution_type,
            in_channels=in_features,
            out_channels=bottleneck_features,
            stride=2 if strided else 1,
            **convolution_params
        )
        self.A1 = mia.make_activation(
            features=bottleneck_features,
            activation_type=activation_type,
            **activation_params
        )
        self.W2 = mic.make_conv_3x3(
            convolution_type=convolution_type,
            in_channels=bottleneck_features,
            out_channels=bottleneck_features,
            **convolution_params
        )
        self.A2 = mia.make_activation(
            features=bottleneck_features,
            activation_type=activation_type,
            **activation_params
        )
        self.W3 = mic.make_conv_1x1(
            convolution_type=convolution_type,
            in_channels=bottleneck_features,
            out_channels=out_features,          
            **convolution_params
        )
        self.A3 = mia.make_activation(
            features=out_features,
            activation_type=activation_type,
            **activation_params
        )
        self.S = torch.nn.Identity() if in_features == out_features\
            else mic.make_conv_1x1(
                    convolution_type=convolution_type,
                    in_channels=in_features,
                    out_channels=out_features,
                    **convolution_params,
                # using a 3x3 conv for shortcut downscaling instead of a 1x1 (used in detectron2 for example)
                ) if not strided else mic.make_conv_3x3(
                    convolution_type=convolution_type,
                    in_channels=in_features,
                    out_channels=out_features,
                    stride=2,
                    **convolution_params,
                )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = self.W3(self.A2(self.W2(self.A1(self.W1(x)))))  # y = W3 * A2(W2 * A1(W1 * x))
        return self.A3(self.S(x) + y)                       # out = A3(S(x) + y)

class PreResBottleneck(Bottleneck):
    def __init__(self,
        convolution_type: str,
        activation_type: str,
        in_features: int,
        out_features: int,
        bottleneck_features: int,
        convolution_params: dict,
        activation_params: dict,
        strided: bool,
    ):
        super(PreResBottleneck, self).__init__(
            convolution_type=convolution_type,
            activation_type=activation_type,
            in_features=in_features,
            out_features=out_features,
            bottleneck_features=bottleneck_features,
            convolution_params=convolution_params,
            activation_params=activation_params,
            strided=strided
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = self.A3(self.W3(self.A2(self.W2(self.A1(self.W1(x))))))  # y = A3(W3 * A2(W2 * A1(W1 * x)))
        return self.S(x) + y                                         # out = S(x) + y

class PreActivBottleneck(torch.nn.Module):
    def __init__(self,
        convolution_type: str,
        activation_type: str,
        in_features: int,
        out_features: int,
        bottleneck_features: int,
        convolution_params: dict,
        activation_params: dict,
        strided: bool,
    ):
        super(PreActivBottleneck, self).__init__()
        self.A1 = mia.make_activation(
            features=in_features,
            activation_type=activation_type,
            **activation_params
        )
        self.W1 = mic.make_conv_1x1(
            convolution_type=convolution_type,
            in_channels=in_features,
            out_channels=bottleneck_features,
            stride=2 if strided else 1,
            **convolution_params
        )
        self.A2 = mia.make_activation(
            features=bottleneck_features,
            activation_type=activation_type,
            **activation_params
        )
        self.W2 = mic.make_conv_3x3(
            convolution_type=convolution_type,
            in_channels=bottleneck_features,
            out_channels=bottleneck_features,
            **convolution_params
        )
        self.A3 = mia.make_activation(
            features=bottleneck_features,
            activation_type=activation_type,
            **activation_params
        )
        self.W3 = mic.make_conv_1x1(
            convolution_type=convolution_type,
            in_channels=bottleneck_features,
            out_channels=out_features,          
            **convolution_params
        )
        self.S = torch.nn.Identity() if in_features == out_features\
            else mic.make_conv_1x1(
                convolution_type=convolution_type,
                in_channels=in_features,
                out_channels=out_features,
                **convolution_params,
            # using a 3x3 conv for shortcut downscaling instead of a 1x1 (used in detectron2 for example)
            ) if not strided else mic.make_conv_3x3(
                convolution_type=convolution_type,
                in_channels=in_features,
                out_channels=out_features,
                stride=2,
                **convolution_params,
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = self.W3(self.A3(self.W2(self.A2(self.W1(self.A1(x)))))) # y = W3 * A3(W2 * A2(W1 * A1(x)))
        return self.S(x) + y                                        # out = x + y



from moai.utils.arguments import assert_path
from moai.data.datasets.common.image2d import load_color_image

import json
import numpy as np
import cv2
import glob
import torch
import os
import typing
import logging
import trimesh

log = logging.getLogger(__name__)

__all__ = ['EHF']

class EHF(torch.utils.data.Dataset):
    def __init__(self,
        root:           str,
        load_jpg:       bool=False,
        # load_scans:     bool=False,
    ) -> None:
        super().__init__()
        assert_path(log, 'EHF root path', root)
        self.root = root
        # self.load_scans = load_scans        
        self.fits = glob.glob(os.path.join(root, 'fits', '*.ply'))
        if load_jpg:
            self.images = glob.glob(os.path.join(root, 'images', '*.jp*g'))
        else:
            self.images = glob.glob(os.path.join(root, 'png', '*.png'))        
        self.focal_length = torch.Tensor([1498.22426237, 1498.22426237]).float()
        self.principal_point = torch.Tensor([790.263706, 578.90334]).float()
        self.camera_translation = torch.Tensor([-0.03609917,  0.43416458,  2.37101226]).float()
        self.camera_rotation = torch.Tensor([-2.98747896,  0.01172457, -0.05704687]).float()
        self.intrinsics = torch.Tensor([
            [1498.22426237, 0.0, 790.263706],
            [0.0, 1498.22426237, 578.90334],
            [0.0, 0.0, 1.0],
        ]).float()
        xform = np.eye(4, dtype=np.float32)
        xform[:3, :3] = cv2.Rodrigues(self.camera_rotation.numpy())[0]
        xform[3, :3] = self.camera_translation.numpy().copy()
        self.extrinsics = torch.from_numpy(xform)        
        
    def __len__(self) -> int:
        return len(self.fits)

    def __getitem__(self, index: int) -> typing.Dict[str, torch.Tensor]:
        fit_filename = self.fits[index]
        img_filename = self.images[index]
        base, ext = os.path.splitext(os.path.basename(img_filename))
        image = load_color_image(img_filename)
        smplx = trimesh.load(fit_filename, process=False)
        keypoints, confidence = self._read_keypoints(os.path.join(
            self.root, 'keypoints', f"{base}_keypoints.json"
        ))
        data = {
            'color': image,
            'keypoints': keypoints,
            'confidence': confidence,
            'vertices': torch.from_numpy(
                np.dot(smplx.vertices, self.extrinsics[:3, :3].numpy().T) + self.camera_translation.numpy()
            ).float(),
            'faces': torch.from_numpy(smplx.faces).int(), #NOTE or long?
            'camera_rotation': self.camera_rotation,
            'camera_translation': self.camera_translation,
            'focal_length': self.focal_length,
            'principal_point': self.principal_point,
            'intrinsics': self.intrinsics,
            'extrinsics': self.extrinsics,
        }
        # if self.load_scans:
        #     scan = trimesh.load(fit_filename.replace('_align.ply', '_scan.obj'))
        #     data['scan_vertices'] = torch.from_numpy(scan.vertices).float()
        #     data['scan_faces'] = torch.from_numpy(scan.faces).int() #NOTE: or long?
        return data

    def _read_keypoints(self, filename: str) -> torch.Tensor:
        with open(filename) as keypoint_file:
            data = json.load(keypoint_file)
        person = data['people'][0]
        body = np.array(person['pose_keypoints_2d'], dtype=np.float32)
        body = body.reshape([-1, 3])
        if True: # load_hands:
            left_hand = np.array(person['hand_left_keypoints_2d'], dtype=np.float32).reshape([-1, 3])
            right_hand = np.array(person['hand_right_keypoints_2d'], dtype=np.float32).reshape([-1, 3])
            body = np.concatenate([body, left_hand, right_hand], axis=0)
        if True: # load_face:
            face = np.array(person['face_keypoints_2d'], dtype=np.float32).reshape([-1, 3])[17: 17 + 51, :]
            contour_keyps = np.array([], dtype=body.dtype).reshape(0, 3)
            if False: # load_face_contour:
                contour_keyps = np.array(person['face_keypoints_2d'], dtype=np.float32).reshape([-1, 3])[:17, :]
            body = np.concatenate([body, face, contour_keyps], axis=0)            
        return torch.from_numpy(body[:, :2]), torch.from_numpy(body[:, 2:3])


from moai.monads.generation.grid.spatial import (
    Grid,
    MultiScaleGrid
)
from moai.monads.generation.grid.spherical import Spherical

__all__ = [
    "Grid",
    "MultiScaleGrid",
    "Spherical",
]


import moai.serve.model as model_server
import moai.serve.optimizer as optimizer_server

import yaml
import toolz
import sys
import hydra
import omegaconf.omegaconf
import logging
import subprocess
import typing
import glob
import os
import zipfile
import toolz

log = logging.getLogger(__name__)

def get_files(
    root:       str,
    folder:     typing.Union[str, typing.Sequence[str]],
    pattern:    str,
) -> typing.Sequence[str]:
    folders = [folder] if isinstance(folder, str) else folder
    folders = [os.path.join(root, f) for f in folders]
    files = []
    for f in folders:
        path = f if os.path.isabs(f) else os.path.join(hydra.utils.get_original_cwd(), f)        
        files += glob.glob(os.path.join(path, pattern))
        files += glob.glob(os.path.join(path, "**", pattern), recursive=True)
    return list(toolz.unique(files))

def dump_zipfile(
    filename:   str,
    root:       str,
    files:      typing.Sequence[str],
):
    with zipfile.ZipFile(filename, "w", zipfile.ZIP_DEFLATED) as z:
        for file in files:                        
            arcname = os.path.join('.', file.replace(root, ''))
            z.write(file, arcname)

def dump_handlers(
    handlers:               omegaconf.omegaconf.DictConfig,
    handler_config:         omegaconf.omegaconf.DictConfig,
):
    empty = omegaconf.DictConfig({})
    with open('pre.yaml', 'w') as f:
        data = omegaconf.OmegaConf.to_container(handlers.get('pre') or empty)
        data = list(map(lambda e: {e.split(':')[0].strip(): e.split(':')[1].strip()}, data))
        yaml.dump({ 'defaults': data }, f, default_style=None, default_flow_style=False)
    with open('post.yaml', 'w') as f:
        data = omegaconf.OmegaConf.to_container(handlers.get('post') or empty)
        data = list(map(lambda e: {e.split(':')[0].strip(): e.split(':')[1].strip()}, data))
        yaml.dump({ 'defaults': data }, f, default_style=None, default_flow_style=False)
    if handler_config is not None:
        with open('handler_overrides.yaml', 'w') as f:
            data = omegaconf.OmegaConf.to_container(handler_config)
            yaml.dump({'handlers': data}, f, default_style=None, default_flow_style=False)
        return True
    return False

def archive(cfg):
    hydra.utils.log.debug(f"Configuration:\n{omegaconf.OmegaConf.to_yaml(cfg, resolve=True)}")    
    omegaconf.omegaconf.OmegaConf.set_struct(cfg, False)
    args = []
    args += ["torch-model-archiver"]
    args += ["--model-name", cfg.archive.name]
    args += ["--version", str(cfg.archive.version)]
    args += ["--handler", optimizer_server.__file__ if cfg.archive.mode == 'fit' else model_server.__file__]
    args += ["--serialized-file", toolz.get_in(['archive', 'ckpt'], cfg) or '']
    yaml_files = get_files(cfg.archive.root, cfg.archive.conf, "*.yaml")
    py_files = get_files(cfg.archive.root, cfg.archive.src, "*.py")
    dump_zipfile("conf.zip", cfg.archive.root, yaml_files)
    dump_zipfile("src.zip", cfg.archive.root, py_files)    
    extra_files = "conf.zip,src.zip,.hydra/overrides.yaml,pre.yaml,post.yaml"
    if dump_handlers(cfg.archive.handlers, cfg.handlers):
        extra_files += ",handler_overrides.yaml"
    args += [f"--extra-files", extra_files]
    # confs = [cfg.archive.conf] if isinstance(cfg.archive.conf, str) else cfg.archive.conf
    # srcs = [cfg.archive.src] if isinstance(cfg.archive.src, str) else cfg.archive.src
    # args += [f"--extra-files", ",".join(map(lambda p: os.path.join(cfg.archive.root, p, ''), confs + srcs))]
    log.info(f"Running: {args}")
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=None)# , shell=True)
    output = proc.communicate()[0].decode('UTF-8')
    if not output:
        log.info(f"Model successfully archived @ {os.getcwd()}")
    else:
        log.error(f"An error has occured while archiving the model:\n{output}")
    
if __name__ == "__main__":  
    config_filename = sys.argv.pop(1) #TODO: argparser integration?
    sys.argv.append("hydra.run.dir=actions/${hydra.job.name}/${now:%Y-%m-%d}/${now:%H-%M-%S}-${archive.name}")
    archive = hydra.main(config_path="conf", config_name=config_filename)(archive)
    archive()


from functools import partial

import torch

class Accuracy(torch.nn.Module):
    def __init__(self,
        topk:       int=1,
    ):
        super(Accuracy, self).__init__()
        self.topk = topk

    def forward(self,
        gt:         torch.Tensor,
        pred:       torch.Tensor,
        weights:    torch.Tensor=None,
        mask:       torch.Tensor=None,
    ) -> torch.Tensor:
        values, indices = torch.topk(
            pred, k=self.topk, dim=1, largest=True, sorted=True
        )
        b = gt.shape[0]
        indices = indices.view(b, self.topk, -1)
        gt = gt.view(b, 1, -1).expand_as(indices)
        correct = torch.eq(gt, indices)
        correct = correct.sum(dim=1).float()
        if weights is not None:
            correct = correct * weights
        if mask is not None:
            correct = correct[mask]
        return correct.mean()

TopAccuracy = partial(Accuracy, topk=1)
Top5Accuracy = partial(Accuracy, topk=5)


from moai.data.datasets.common.image2d import load_color_image
from collections.abc import Callable

import typing
import torch
import logging
import cv2

__all__ = [
    'ImageFileInput',
    'ImageFileOutput'
]

log = logging.getLogger(__name__)

class ImageFileInput(Callable):
    def __init__(self,
        input_key:          str='color',
        output_key:         str='color',
    ) -> None:
        super().__init__()
        self.input_key = input_key
        self.output_key = output_key

    def __call__(self, 
        data:   typing.Mapping[str, typing.Any],
        device: torch.device,
    ) -> torch.Tensor:
        path = data.get(self.input_key)
        log.info(f"Loading image file [key: {self.input_key}] @ {path}")
        return {self.output_key: load_color_image(path).unsqueeze(0).to(device) }

class ImageFileOutput(Callable):
    def __init__(self,
        input_key:          str='color',
    ) -> None:
        super().__init__()
        self.input_key = input_key

    def __call__(self, 
        data:   typing.Mapping[str, torch.Tensor],
        json:   typing.Mapping[str, typing.Any],
    ) -> typing.Dict[str, torch.Tensor]:
        image = data.get(self.input_key)
        image = image.detach().cpu().numpy()
        for img, kvp in zip(image, json):
            filename = kvp['body'].get(self.input_key)
            log.info(f"Saving tensor [key: {self.input_key}] @ {filename}")
            cv2.imwrite(filename, img.transpose(1, 2, 0))
        return { self.input_key: 'Success' } #NOTE: Object type should be JSON serializable


from mesh_intersection.bvh_search_tree import BVH
from mesh_intersection.filter_faces import FilterFaces
import mesh_intersection.loss as collisions_loss
#TODO: try/catch imports with error message

import torch
import logging
import os
import pickle
import typing

#NOTE: code from https://github.com/vchoutas/torch-mesh-isect

__all__ = ['Interpenetration']

log = logging.getLogger(__name__)

class Interpenetration(torch.nn.Module):
    def __init__(self,
        max_collisions:     int=128,
        df_cone_height:     float=0.0001,
        penalize_outside:   bool=True,
        point2plane:        bool=False,
        part_segmentation:  str='',
        ignore_part_pairs:  typing.Sequence[str]=["9,16", "9,17", "6,16", "6,17", "1,2", "12,22"],
    ):
        super(Interpenetration, self).__init__()
        if not torch.cuda.is_available():
            log.error('The interpenetration loss cannot be used without any CUDA devices.')
        self.search_tree = BVH(max_collisions=max_collisions)
        self.pen_distance = collisions_loss.DistanceFieldPenetrationLoss(
            sigma=df_cone_height, point2plane=point2plane,
            vectorized=True, penalize_outside=penalize_outside
        )
        if part_segmentation:
            if not os.path.exists(part_segmentation):
                log.error(f"Part segmentation file ({part_segmentation}) given to interpenetration loss does not exist.")            
            part_segmentation = os.path.expandvars(part_segmentation)
            with open(part_segmentation, 'rb') as faces_parents_file:
                face_segm_data = pickle.load(faces_parents_file, encoding='latin1')
            faces_segm = face_segm_data['segm']
            faces_parents = face_segm_data['parents']            
            self.filter_faces = FilterFaces(# Create the module used to filter invalid collision pairs
                faces_segm=faces_segm, faces_parents=faces_parents,
                ign_part_pairs=ignore_part_pairs
            )
        else:
            self.filter_faces = torch.nn.Identity()

    def forward(self, 
        vertices:   torch.Tensor,
        faces:      torch.Tensor,
    ) -> torch.Tensor:
        b = vertices.shape[0]
        triangles = torch.index_select(vertices, 1, faces.view(-1)).view(b, -1, 3, 3)
        with torch.no_grad():
            collision_idxs = self.search_tree(triangles)        
        collision_idxs = self.filter_faces(collision_idxs) # Remove unwanted collisions
        return self.pen_distance(triangles, collision_idxs) \
            if collision_idxs.ge(0).sum().item() > 0\
                else torch.scalar_tensor(0.0).to(vertices)


import hydra
import sys
import omegaconf.omegaconf
import logging
import typing
import tablib

log = logging.getLogger(__name__)

def assign(cfg: omegaconf.DictConfig, attr: str) -> typing.Union[typing.Any]:
    return getattr(cfg, attr) if hasattr(cfg, attr) else None

def evaluate(cfg):
    hydra.utils.log.debug(f"Configuration:\n{omegaconf.OmegaConf.to_yaml(cfg, resolve=True)}")
    with open("config_resolved.yaml", 'w') as f:
        f.write(omegaconf.OmegaConf.to_yaml(cfg, resolve=True, sort_keys=False))
    with open("config.yaml", 'w') as f:
        f.write(omegaconf.OmegaConf.to_yaml(cfg, resolve=False, sort_keys=False))
    engine = hydra.utils.instantiate(cfg.engine)
    model = hydra.utils.instantiate(cfg.model,
        data=cfg.data,
        visualization=assign(cfg, "visualization"),
        export=assign(cfg, "export"),    
    )
    for name, remodel in (assign(cfg, "remodel") or {}).items():
        hydra.utils.instantiate(remodel)(model)
    model.initialize_parameters()
    tester = hydra.utils.instantiate(cfg.tester, 
        logging=assign(cfg, "logging")
    )
    log.info("Evaluation started.")
    results = tester.run(model)[0] #TODO: need to support other ways of logging average results
    ds = tablib.Dataset(results.values(), headers=results.keys())
    with open(cfg.experiment.name + "_test_average.csv", 'a', newline='') as f:
        f.write(ds.export('csv'))
    log.info("Evaluation completed.")

if __name__ == "__main__":
    # os.environ['HYDRA_FULL_ERROR'] = '1'
    config_filename = sys.argv.pop(1) #TODO: argparser integration?
    sys.argv.append("hydra.run.dir=actions/${hydra.job.name}/${now:%Y-%m-%d}/${now:%H-%M-%S}-${experiment.name}")    
    evaluate = hydra.main(config_path="conf", config_name=config_filename)(evaluate)
    evaluate()


from moai.data.datasets.generic.structured_images import StructuredImages
from moai.data.datasets.generic.npz import Npz
from moai.data.datasets.generic.dummy import (
    Empty, 
    Dummy,
)

__all__ = [
    "StructuredImages",
    "Empty",
    "Dummy",
    "Npz",
]


import cv2
import torch
import torchvision
import typing
import numpy as np

def load_color_image(
    filename:       str,    
    output_space:   str='norm', # ['norm', 'ndc', 'pixel']
) -> torch.Tensor:
    # img = torchvision.io.read_image(filename)
    img = torch.from_numpy(
        cv2.imread(filename).transpose(2, 0, 1)
    ).flip(dims=[0])
    
    if output_space == 'norm':
        img = img / 255.0
    return img

def load_depth_image(
    filename:   str,
    scale:      float=1.0,
) -> torch.Tensor:
    depth = cv2.imread(filename, flags=cv2.IMREAD_ANYDEPTH)
    return torch.from_numpy(depth.astype(np.float32)).unsqueeze(0) * scale

def load_normal_image(
    filename:   str,
    rotation:   typing.List[float]=[],
) -> torch.Tensor:
    pass

def load_mask_image(
    filename:   str,
) -> torch.Tensor:
    img = torch.from_numpy(
        cv2.imread(filename).transpose(2, 0, 1)
    )
    return img[0:1, ...] / 255.0

def load_binary_image(
    filename:   str,
) -> torch.Tensor:
    img = torch.from_numpy(
        cv2.imread(filename).transpose(2, 0, 1)
    )
    return img[0:1, ...].float()



from moai.parameters.optimization.optimizers.lookahead import Lookahead as Outer

import torch
import omegaconf.omegaconf
import hydra.utils as hyu
import typing

__all__ = ['Lookahead']

#NOTE: modified from https://github.com/alphadl/lookahead.pytorch

class Lookahead(object):
    """Implements the Lookahead algorithm.

    - **Paper**: [Lookahead Optimizer: k steps forward, 1 step back](https://arxiv.org/pdf/1907.08610.pdf)
    - **Implementation**: [GitHub @ alphadl](https://github.com/alphadl/lookahead.pytorch)

    """
    def __init__(self,
        parameters: typing.Iterator[torch.nn.Parameter],
        optimizers: omegaconf.DictConfig,
        k: int=5,
        alpha: float=0.5
    ):
        self.optimizers = [
            Outer(
                optimizer=hyu.instantiate(optimizers[0], parameters),
                k=k, alpha=alpha
            )
        ]



import pytorch_lightning
from moai.validation.collection import Metrics
from moai.validation.metrics.common.classified import (
    TruePositive,
    TrueNegative,
    FalsePositive,
    FalseNegative,
)
from pytorch_lightning import Callback
from collections import Iterable

import torch
import omegaconf.omegaconf
import typing
import logging
import toolz
import functools
import numpy as np

log = logging.getLogger(__name__)

__all__ = ['Classification']

def _safe_get_list(
    list: typing.Sequence[typing.Any], 
    index: int=0,
) -> typing.Any:
    return next(iter(list[index:index+1]), None)

def _binary(
    classified:     torch.Tensor,
    index:          int,
):
    return classified[index]

def _total(
    classified:     torch.Tensor,
):
    return torch.sum(classified)

class Classification(Metrics, Callback):
    execs: typing.List[typing.Callable] = []

    _MODES_ = {
        'binary': _binary,
        'total': _total,
    }

    def __init__(self, 
        gt:         typing.Sequence[str]=['label'],
        pred:       typing.Sequence[str]=['prediction'],
        weights:    typing.Sequence[str]=[],
        mask:       typing.Sequence[str]=[],
        # classes:    typing.Union[str, typing.Sequence[str]]='',
        mode:       typing.Union[int, str]='total',
        metrics:    omegaconf.DictConfig={},
    ):
        super(Classification, self).__init__(
            metrics, 
            **dict((k, {
                'gt': gt,
                'pred': pred,
                'weights': weights,
                'mask': mask
            }) for k in metrics.keys())
        )
        self.valid = True
        self.gt, self.pred, self.weights, self.mask = gt, pred, weights or [], mask or []
        # if isinstance(classes, str): #TODO: add loading cases (e.g. np/json files)
        #     with open(classes) as f:
        #         self.classes = [line.rstrip() for line in f]
        # elif isinstance(classes, Iterable):
        #     self.classes = classes
        # elif isinstance(classes, int):
        #     self.classes = ['' for _ in range(classes)]
        # else:
        #     log.error(f"The 'classes' parameter can only be one of [filepath, string list, int], instead it is {type(classes)}.")
        self.mode = Classification._MODES_[mode] if isinstance(mode, str) and mode == 'total'\
            else functools.partial(Classification._MODES_['binary'], index=mode)
        self.measures = { 'TP':0, 'TN': 0, 'FP': 0, 'FN': 0 }
        self.TP = TruePositive()
        self.TN = TrueNegative()
        self.FP = FalsePositive()
        self.FN = FalseNegative()
        self.metrics = None

    def forward(self,
        tensors: typing.Dict[str, torch.Tensor]
    ) -> typing.Dict[str, torch.Tensor]:
        metrics = { }
        for exe in self.execs:
            exe(tensors, metrics)
        returned = { }
        for k, m in metrics.items():
            returned[f'{k}'] = m
        for i, (gt, pred) in enumerate(zip(self.gt, self.pred)):
            w = _safe_get_list(self.weights, i)
            m = _safe_get_list(self.mask, i)
            for f in self.measures.keys():
                returned[f] = self.mode(
                    getattr(self, f)(
                        gt=tensors[gt],
                        pred=tensors[pred],
                        weights=tensors[w] if w else None,
                        mask=tensors[m] if m else None
                    )
                )
        if self.metrics:
            returned = toolz.merge(returned, self.metrics)
            self.metrics = None
        return returned

    def on_sanity_check_start(self, 
        trainer:    pytorch_lightning.Trainer,
        model:      pytorch_lightning.LightningModule,
    ) -> None:
        self.valid = False

    def on_sanity_check_end(self, 
        trainer:    pytorch_lightning.Trainer,
        model:      pytorch_lightning.LightningModule,
    ) -> None:
        self.valid = True

    # def on_validation_epoch_start(self, trainer, model):
    #     pass

    # def on_validation_batch_start(self, 
    #     trainer, model, batch, batch_idx, dataloader_idx
    # ):
    #     pass

    def on_validation_batch_end(self, 
        trainer:        pytorch_lightning.Trainer,
        model:          pytorch_lightning.LightningModule,
        outputs:        dict,
        batch:          dict,
        batch_idx:      int,
        dataloader_idx: int,
    ) -> None:
        if self.valid:
            self.measures['TP'] += int(outputs['TP'])
            self.measures['TN'] += int(outputs['TN'])
            self.measures['FP'] += int(outputs['FP'])
            self.measures['FN'] += int(outputs['FN'])

    # def on_validation_epoch_end(self, trainer, model):
    #     pass

    def on_validation_start(self, 
        trainer:    pytorch_lightning.Trainer,
        model:      pytorch_lightning.LightningModule,
    ) -> None:
        self.measures = { 'TP': 0, 'TN': 0, 'FP': 0, 'FN': 0 }

    def on_validation_end(self, 
    # def on_validation_epoch_end(self,     
        trainer:    pytorch_lightning.Trainer,
        model:      pytorch_lightning.LightningModule,
    ) -> None:
        if not self.valid:
            return
        tp = np.float32(self.measures['TP'])
        tn = np.float32(self.measures['TN'])
        fp = np.float32(self.measures['FP'])
        fn = np.float32(self.measures['FN'])
        prec = tp / (tp + fp)
        recall = tp / (tp + fn)
        specificity = tn / (tn + fp)
        sensitivity = tp / (tp + fn)
        neg_pred_value = tn / (tn + fn)
        threat_score = tp / (tp + fn + fp)
        f1 = (2 * prec * recall) / (prec + recall)
        metrics = {
            'precision': prec,
            'recall': recall,
            'specificity': specificity,
            'negative_predictive_value': neg_pred_value,
            'threat_score': threat_score,
            'F1': f1,
        }
        model.log_dict(metrics, prog_bar=True, logger=False, on_epoch=True, sync_dist=True)
        log_metrics = toolz.keymap(lambda k: f"val_{k}", metrics)
        model.log_dict(log_metrics, prog_bar=False, logger=True, on_epoch=True, sync_dist=True)
        self.metrics = metrics

        


import functools
import torch
import numpy
from matplotlib import cm
from matplotlib.colors import Colormap

# import moai.utils.color.turbo @NOTE check if colormap turbo works with Matplotlib >= 3.3.0

__all__ = [
    "get_colormap",
    "COLORMAPS"
]

def _matplotlib_colormap(
    colormap: Colormap,
    tensor: torch.Tensor
) -> numpy.array:
    data = tensor.cpu().detach().numpy() if tensor.is_cuda else tensor.detach().numpy()
    return colormap(data).squeeze(1).transpose(0, 3, 1, 2)[:, :3, :, :]

jet = functools.partial(_matplotlib_colormap, cm.get_cmap('jet'))
magma = functools.partial(_matplotlib_colormap, cm.get_cmap('magma'))
inferno = functools.partial(_matplotlib_colormap, cm.get_cmap('inferno'))
plasma = functools.partial(_matplotlib_colormap, cm.get_cmap('plasma'))
seismic = functools.partial(_matplotlib_colormap, cm.get_cmap('seismic'))
viridis = functools.partial(_matplotlib_colormap, cm.get_cmap('viridis'))
viridis_r = functools.partial(_matplotlib_colormap, cm.get_cmap('viridis_r'))
gray = functools.partial(_matplotlib_colormap, cm.get_cmap('gray'))
cividis = functools.partial(_matplotlib_colormap, cm.get_cmap('cividis'))
bone = functools.partial(_matplotlib_colormap, cm.get_cmap('bone'))
bone_r = functools.partial(_matplotlib_colormap, cm.get_cmap('bone_r'))
turbo = functools.partial(_matplotlib_colormap, cm.get_cmap('turbo'))
turbo_r = functools.partial(_matplotlib_colormap, cm.get_cmap('turbo_r'))

COLORMAPS = {
    'jet': jet,
    'magma': magma,
    'inferno': inferno,
    'plasma': plasma,
    'seismic': seismic,
    'viridis': viridis,
    'viridis_r': viridis_r,
    'gray': gray,
    'cividis': cividis,
    'turbo': turbo,
    'turbo_r': turbo_r,
    'bone': bone,
    'bone_r': bone_r,
}

def get_colormap(name:str) -> Colormap:
    return COLORMAPS[name] if name in COLORMAPS.keys() else COLORMAPS.items()[0][1]


    